# -*-coding:utf-8-*-
import ast
import uuid
import json
from enum import Enum
from typing import List, Dict

from ultipa.structs.Process import Process

from ultipa.structs.Privilege import Privilege, PrivilegeLevel
import datetime
from ultipa.structs import Index, Policy, User, Stats
from ultipa.structs.DBType import DBType
from ultipa.structs.Algo import Algo
from ultipa.structs.HDC import HDCGraph, HDCSyncType
from ultipa.structs.BaseModel import BaseModel
from ultipa.structs.Edge import Edge
from ultipa.structs.EntityRow import EntityRow
from ultipa.structs.Node import Node, NodeAlias
from ultipa.structs.Path import Path, PathAlias
from ultipa.structs.Projection import Projection
from ultipa.structs.Property import Property
from ultipa.structs.PropertyPrivilege import PropertyPrivilege, PropertyPrivilegeElement
from ultipa.structs.PropertyType import *
from ultipa.structs.ResultType import ResultType
from ultipa.structs.Retry import Retry
from ultipa.structs.Schema import Schema
from ultipa.structs.GraphSet import GraphSet
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from ultipa.structs.Graph import Graph, GraphAlias
from ultipa.utils.convert import convertToListAnyObject, convertTableToDict, convertToTask, converToJob, \
    converToProjection
from ultipa.utils.errors import ParameterException


# class TruncateType:
# 	NODES = 'nodes'
# 	EDGES = 'edges'


class DirectionType:
    left = 'left'
    right = 'right'


class TaskStatus:
    TASK_WAITING = 0
    TASK_COMPUTING = 1
    TASK_WRITEBACKING = 2
    TASK_DONE = 3
    TASK_FAILED = 4
    TASK_STOP = 5


TaskStatusString = {
    TaskStatus.TASK_WAITING: "TASK_WAITING",
    TaskStatus.TASK_COMPUTING: "TASK_COMPUTING",
    TaskStatus.TASK_WRITEBACKING: "TASK_WRITEBACKING",
    TaskStatus.TASK_DONE: "TASK_DONE",
    TaskStatus.TASK_FAILED: "TASK_FAILED",
    TaskStatus.TASK_STOP: "TASK_STOP"
}


class ALGO_RETURN_TYPE:
    ALGO_RETURN_REALTIME = 1
    ALGO_RETURN_WRITE_BACK = 2
    ALGO_RETURN_VISUALIZATION = 4


class ErrorCode(Enum):
    SUCCESS = ultipa_pb2.SUCCESS
    FAILED = ultipa_pb2.FAILED
    PARAM_ERROR = ultipa_pb2.PARAM_ERROR
    BASE_DB_ERROR = ultipa_pb2.BASE_DB_ERROR
    ENGINE_ERROR = ultipa_pb2.ENGINE_ERROR
    SYSTEM_ERROR = ultipa_pb2.SYSTEM_ERROR
    SYNTAX_ERROR = ultipa_pb2.SYNTAX_ERROR
    RAFT_REDIRECT = ultipa_pb2.RAFT_REDIRECT
    RAFT_LEADER_NOT_YET_ELECTED = ultipa_pb2.RAFT_LEADER_NOT_YET_ELECTED
    RAFT_LOG_ERROR = ultipa_pb2.RAFT_LOG_ERROR
    # UQL_ERROR = ultipa_pb2.UQL_ERROR
    NOT_RAFT_MODE = ultipa_pb2.NOT_RAFT_MODE
    RAFT_NO_AVAILABLE_FOLLOWERS = ultipa_pb2.RAFT_NO_AVAILABLE_FOLLOWERS
    RAFT_NO_AVAILABLE_ALGO_SERVERS = ultipa_pb2.RAFT_NO_AVAILABLE_ALGO_SERVERS
    PERMISSION_DENIED = ultipa_pb2.PERMISSION_DENIED
    DUPLICATE_ID = ultipa_pb2.DUPLICATE_ID

    UNKNOW_ERROR = 1000


class CodeMap:
    _codeMap = {
        ultipa_pb2.SUCCESS: ErrorCode.SUCCESS,
        ultipa_pb2.FAILED: ErrorCode.FAILED,
        ultipa_pb2.PARAM_ERROR: ErrorCode.PARAM_ERROR,
        ultipa_pb2.BASE_DB_ERROR: ErrorCode.BASE_DB_ERROR,
        ultipa_pb2.ENGINE_ERROR: ErrorCode.ENGINE_ERROR,
        ultipa_pb2.SYSTEM_ERROR: ErrorCode.SYSTEM_ERROR,
        ultipa_pb2.SYNTAX_ERROR: ErrorCode.SYNTAX_ERROR,
        ultipa_pb2.RAFT_REDIRECT: ErrorCode.RAFT_REDIRECT,
        ultipa_pb2.RAFT_LEADER_NOT_YET_ELECTED: ErrorCode.RAFT_LEADER_NOT_YET_ELECTED,
        ultipa_pb2.RAFT_LOG_ERROR: ErrorCode.RAFT_LOG_ERROR,
        # ultipa_pb2.UQL_ERROR : UQL_ERROR.name,
        ultipa_pb2.NOT_RAFT_MODE: ErrorCode.NOT_RAFT_MODE,
        ultipa_pb2.RAFT_NO_AVAILABLE_FOLLOWERS: ErrorCode.RAFT_NO_AVAILABLE_FOLLOWERS,
        ultipa_pb2.RAFT_NO_AVAILABLE_ALGO_SERVERS: ErrorCode.RAFT_NO_AVAILABLE_ALGO_SERVERS,
        ultipa_pb2.PERMISSION_DENIED: ErrorCode.PERMISSION_DENIED,
        ultipa_pb2.DUPLICATE_ID: ErrorCode.DUPLICATE_ID,
        1000: ErrorCode.UNKNOW_ERROR

    }

    def getCode(codeKey: str):
        return CodeMap._codeMap[codeKey]


class FollowerRole:
    ROLE_UNSET = ultipa_pb2.ROLE_UNSET
    ROLE_READABLE = ultipa_pb2.ROLE_READABLE
    ROLE_ALGO_EXECUTABLE = ultipa_pb2.ROLE_ALGO_EXECUTABLE


class RaftPeerInfo:
    def __init__(self, host, status=None, isLeader=False, isAlgoExecutable=False, isFollowerReadable=False,
                 isUnset=False):
        self.host = host
        self.status = status
        self.isLeader = isLeader
        self.isAlgoExecutable = isAlgoExecutable
        self.isFollowerReadable = isFollowerReadable
        self.isUnset = isUnset


class ClusterInfo:
    def __init__(self, redirect: str, raftPeers: List[RaftPeerInfo], leader: RaftPeerInfo = None):
        self.redirect = redirect
        self.leader = leader
        self.raftPeers = raftPeers


class Status:
    '''
        Data class for Status
    '''

    def __init__(self, message: str = None, code: ErrorCode = None):
        self.code = code
        self.message = message.strip('\n')


class Schemas:
    def __init__(self, schema: List[Schema] = None, totalNodes: int = None, totalEdges: int = None):
        self.schema = schema
        self.totalNodes = totalNodes
        self.totalEdges = totalEdges


class NodeEntityTable:
    def __init__(self, schemas: List[object], nodeRows: List[EntityRow] = None):
        self.schemas = schemas
        if nodeRows is None:
            nodeRows = []
        self.nodeRows = nodeRows

    def __del__(self):
        pass


class EdgeEntityTable:
    def __init__(self, schemas: List[object], edgeRows: List[EntityRow] = None):
        self.schemas = schemas
        if edgeRows is None:
            edgeRows = []
        self.edgeRows = edgeRows

    def __del__(self):
        pass


class EdgeAlias:
    def __init__(self, alias: str, edges: List[Edge]):
        self.alias = alias
        self.edges = edges


class Attr(BaseModel):
    '''
            Data class for attr
    '''

    def __init__(self, name: str = None,
                 values: List[any] = None,
                 resultType: ResultType = None,
                 propertyType: UltipaPropertyType = None):
        self.name = name
        self.values = values
        self.resultType = resultType
        self.propertyType = propertyType


class AttrNode:
    def __init__(self, alias: str, values: List[List[Node]], type: ResultType = None, type_desc: str = None):
        self.name = alias
        self.values = values
        self.type = type
        self.type_desc = type_desc


class AttrEdge:
    def __init__(self, alias: str, values: List[List[Edge]], type: ResultType = None, type_desc: str = None):
        self.name = alias
        self.values = values
        self.type = type
        self.type_desc = type_desc


class AttrPath:
    def __init__(self, alias: str, values: List[List[Path]], type: ResultType = None, type_desc: str = None):
        self.name = alias
        self.values = values
        self.type = type
        self.type_desc = type_desc


class UltipaAttr:

    def __init__(self, type: UltipaPropertyType, values: any, has_attr_data: bool = False,
                 has_ultipa_data: bool = False,
                 type_desc: any = None):
        self.values = values
        self.type = type
        self.type_desc = type_desc
        self.has_attr_data = has_attr_data
        self.has_ultipa_data = has_ultipa_data


class AttrNewAlias:
    def __init__(self, alias: str, attr: UltipaAttr):
        self.alias = alias
        self.attr = attr


class Alias:
    '''
        Data class for Alias
    '''

    def __init__(self, alias: str = None, type: ResultType = None):
        self.alias = alias
        self.type = type


class Table(BaseModel):
    '''
        Data class for table.
    '''

    def __init__(self, name: str = None, headers: List[dict] = None, rows: List[List[any]] = None):
        self.name = name
        self.rows = rows
        self.headers = headers

    def getHeaders(self):
        return self.headers

    def getRows(self):
        return self.rows

    def getName(self):
        return self.name

    def toKV(self) -> List[Dict]:
        return convertTableToDict(self.rows, self.headers)

    def headerToDicts(self) -> List[Dict]:
        return convertTableToDict(self.rows, self.headers)


class ArrayAlias:
    def __init__(self, alias: str, elements):
        self.alias = alias
        self.elements = elements


class Exta(BaseModel):
    '''
            Data class for exta
    '''

    def __init__(self, author: str = None, detail: str = None, name: str = None, version: str = None):
        self.author = author
        self.detail = detail
        self.name = name
        self.version = version


class PlanNode:
    '''

        Data class for PlanNode

    '''

    def __init__(self, alias: str = None, childrenNum: int = None, uql: str = None, infos: str = None):
        self.alias = alias
        self.children_num = childrenNum
        self.uql = uql
        self.infos = infos


class ExplainPlan:
    '''
        Data class for ExplainPlan

    '''

    def __init__(self, planNodes: List[PlanNode] = None):
        self.planNodes = planNodes


class DataItem(BaseModel):
    '''
        Data class for DataItem

    '''

    def __init__(self, alias: str, entities: any, type: ResultType):
        self.alias = alias
        self.entities = entities
        self.type = type

    def asNodes(self) -> List[Node]:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_NODE):
            error = f"DataItem {self.alias} is not Type Node"
            raise ParameterException(error)
        return self.entities

    def asFirstNode(self) -> Node:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_NODE):
            error = f"DataItem {self.alias} is not Type Node"
            raise ParameterException(error)
        return self.entities[0] if len(self.entities) > 0 else None

    def asEdges(self) -> List[Edge]:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_EDGE):
            error = f"DataItem {self.alias} is not Type Edge"
            raise ParameterException(error)
        return self.entities

    def asFirstEdge(self) -> Edge:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_EDGE):
            error = f"DataItem {self.alias} is not Type Edge"
            raise ParameterException(error)
        return self.entities[0] if len(self.entities) > 0 else None

    def asAttr(self) -> Attr:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_ATTR):
            error = f"DataItem {self.alias} is not Type Attribute list"
            raise ParameterException(error)

        return self.entities

    def asAttrNodes(self) -> AttrNode:
        return self.asAttr()

    def asAttrEdges(self) -> AttrEdge:
        return self.asAttr()

    def asAttrPaths(self) -> AttrPath:
        return self.asAttr()

    def asPaths(self) -> List[Path]:
        graphdata = self.entities
        for data in graphdata.paths:
            data.nodes = {nodeuuidsdata: graphdata.nodes[nodeuuidsdata] for nodeuuidsdata in data.nodeUuids}
            data.edges = {edgeuuidsdata: graphdata.edges[edgeuuidsdata] for edgeuuidsdata in data.edgeUuids}
        return graphdata.paths

    def asTable(self) -> Table:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
            error = f"DataItem {self.alias} is not Type Table"
            raise ParameterException(error)
        return self.entities

    def asSchemas(self) -> List[Schema]:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
            error = f"DataItem {self.alias} is not Type Table"
            raise ParameterException(error)
        alias = self.entities.getName()
        if alias.startswith("_node"):
            type = "node"
            dbType = DBType.DBNODE
        elif alias.startswith("_edge"):
            type = "edge"
            dbType = DBType.DBEDGE

        else:
            type = None
            dbType = None

        headers = self.entities.getHeaders()
        rows = self.entities.getRows()
        tableListDict = convertTableToDict(rows, headers)
        REPLACE_KEYS = {
            "totalNodes": "total",
            "totalEdges": "total",
        }
        BOOL_KEYS = ["index", "lte"]
        JSON_KEYS = ["properties"]
        convert2Int = ["totalNodes", "totalEdges"]
        responseKeyFormat = ResponseKeyFormat(keyReplace=REPLACE_KEYS, boolKeys=BOOL_KEYS, jsonKeys=JSON_KEYS,
                                              convert2Int=convert2Int)
        dataList = responseKeyFormat.changeKeyValue(tableListDict)
        schemaList = []

        def none_k(data_none: str):
            return '' if data_none is None else data_none

        for data in dataList:
            responseKeyFormat = ResponseKeyFormat(boolKeys=BOOL_KEYS)
            properList = responseKeyFormat.changeKeyValue(data.get("properties"))
            propertyList = []
            for propo in properList:
                prop = Property(name=none_k(propo.get("name")),
                                description=none_k(propo.get("description")),
                                type=none_k(propo.get("type")),
                                lte=propo.get("lte"),
                                schema=data.get("name"),
                                read=propo.get("read"),
                                write=propo.get("write"),
                                encrypt=none_k(propo.get("encrypt")
                                               ))

                prop.type = prop.getPropertyTypeByString(prop.type)
                propertyList.append(prop)

            schemaList.append(
                Schema(name=data.get("name"), description=data.get("description"),
                       properties=propertyList,
                       dbType=dbType,
                       id=int(data.get('id')) if data.get('id').isdigit() else None,
                       total=int(data.get("total")) if data.get("total") and data.get("total") != '' else data.get(
                           "total")))

        return schemaList

    def asProperties(self) -> List[Property]:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
            error = f"DataItem {self.alias} is not Type Table"
            raise ParameterException(error)
        headers = self.entities.getHeaders()
        rows = self.entities.getRows()
        table_rows_dict = convertTableToDict(rows, headers)
        BOOL_KEYS = ["lte", "read", "write"]
        responseKeyFormat = ResponseKeyFormat(boolKeys=BOOL_KEYS)
        dataList = responseKeyFormat.changeKeyValue(table_rows_dict)

        def none_k(data_none: str):
            return '' if data_none is None else data_none

        propertyList = []
        for propo in dataList:
            prop = Property(name=none_k(propo.get("name")),
                            description=none_k(propo.get("description")),
                            type=none_k(propo.get("type")),
                            lte=propo.get("lte"),
                            schema=propo.get("schema"),
                            read=propo.get("read"),
                            write=propo.get("write"),
                            encrypt=none_k(propo.get("encrypt")
                                           ))

            prop.type = prop.getPropertyTypeByString(prop.type)
            propertyList.append(prop)

        return propertyList

    def asGraphSets(self) -> List[GraphSet]:
        REPLACE_KEYS = {
            "graph": "name",
        }
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
            error = f"DataItem {self.alias} is not Type Table"
            raise ParameterException(error)
        headers = self.entities.getHeaders()
        rows = self.entities.getRows()
        table_rows_dict = convertTableToDict(rows, headers)
        responseKeyFormat = ResponseKeyFormat(keyReplace=REPLACE_KEYS)
        data = responseKeyFormat.changeKeyValue(table_rows_dict)
        data = [GraphSet(id=res.id, description=res.description, name=res.name,
                         totalNodes=int(res.total_nodes) if hasattr(res,
                                                                    'total_nodes') and res.total_nodes.isdigit() else 0,
                         totalEdges=int(res.total_edges) if hasattr(res,
                                                                    'total_edges') and res.total_edges.isdigit() else 0,
                         slotNum=int(res.slot_num) if res.slot_num.isdigit() else 0,
                         shards=res.shards.split(','),
                         partitionBy=res.partition_by,
                         status=res.status
                         ) for res in convertToListAnyObject(data)]
        return data

    def asGraph(self) -> Graph:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.RESULT_TYPE_GRAPH:
            error = f"DataItem {self.alias} is not Type Graph"
            raise ParameterException(error)
        return Graph(paths=self.entities.paths, nodes=self.entities.nodes, edges=self.entities.edges)

    def asAlgos(self) -> List[Algo]:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
            error = f"DataItem {self.alias} is not Type Table"
            raise ParameterException(error)
        headers = self.entities.getHeaders()
        rows = self.entities.getRows()
        table_rows_dict = convertTableToDict(rows, headers)
        algo_list = []
        for paramDict in table_rows_dict:
            desc = json.loads(paramDict.get("description"))
            algoObj = Algo(name=paramDict.get("name"),
                           version=desc.get('version'),
                           type=paramDict.get("type"),
                           params=paramDict.get("params"),
                           description=paramDict.get("description"),
                           canRollback=paramDict.get("can_rollback"),
                           configContext=paramDict.get("config_context"),
                           writeSupportType=paramDict.get('write_support_type')
                           )

            algo_list.append(algoObj)
        return algo_list

    def asHDCGraphs(self) -> List[HDCGraph]:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
            error = f"DataItem {self.alias} is not Type Table"
            raise ParameterException(error)
        headers = self.entities.getHeaders()
        rows = self.entities.getRows()
        table_rows_dict = convertTableToDict(rows, headers)
        algo_list = []
        for paramDict in table_rows_dict:
            algoObj = HDCGraph(name=paramDict.get("name"),
                               graphName=paramDict.get("graph_name"),
                               status=paramDict.get("status"),
                               stats=paramDict.get('stats'),
                               isDefault=paramDict.get("is_default"),
                               hdcServerName=paramDict.get('hdc_server_name'),
                               hdcServerStatus=paramDict.get('hdc_server_status'),
                               config=paramDict.get('config')
                               )

            algo_list.append(algoObj)
        return algo_list

    def asExtas(self) -> List[Exta]:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
            error = f"DataItem {self.alias} is not Type Table"
            raise ParameterException(error)

        headers = self.entities.getHeaders()
        rows = self.entities.getRows()
        table_rows_dict = convertTableToDict(rows, headers)
        exta_list = []
        for data in table_rows_dict:
            exta = Exta(author=data.get('author'),
                        detail=data.get('detail'),
                        name=data.get('name'),
                        version=data.get('version'))

            exta_list.append(exta)

        return exta_list

    def asIndexes(self) -> List[Index]:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
            error = f"DataItem {self.alias} is not Type Table"
            raise ParameterException(error)
        headers = self.entities.getHeaders()
        rows = self.entities.getRows()
        table_rows_dict = convertTableToDict(rows, headers)
        index_list = []
        for data in table_rows_dict:
            index = Index(
                id=data.get('id'),
                dbType=data.get('db_type'),
                name=data.get('name'),
                properties=data.get('properties'),
                schema=data.get('schema'),
                size=data.get('size'),
                status=data.get('status'))

            index_list.append(index)

        return index_list

    def asPrivileges(self) -> List[Privilege]:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return None
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
            error = f"DataItem {self.alias} is not Type Table"
            raise ParameterException(error)
        headers = self.entities.getHeaders()
        rows = self.entities.getRows()
        table_rows_dict = convertTableToDict(rows, headers)
        privilege_list = []
        for data in table_rows_dict:
            sysPrivileges = [Privilege(name=sysdata, level=PrivilegeLevel.SYSTEM) for sysdata in
                             ast.literal_eval(data.get('systemPrivileges'))]
            privilege_list.extend(sysPrivileges)
            graPrivileges = [Privilege(name=gradata, level=PrivilegeLevel.GRAPH) for gradata in
                             ast.literal_eval(data.get('graphPrivileges'))]
            privilege_list.extend(graPrivileges)
        return privilege_list

    def asPolicies(self) -> List[Policy]:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
            error = f"DataItem {self.alias} is not Type Table"
            raise ParameterException(error)
        headers = self.entities.getHeaders()
        rows = self.entities.getRows()
        table_rows_dict = convertTableToDict(rows, headers)
        policy_list = []
        for data in table_rows_dict:
            proPrivilegesData = json.loads(data.get('propertyPrivileges'))
            policy = Policy(
                name=data.get('name'),
                graphPrivileges=json.loads(data.get('graphPrivileges')),
                systemPrivileges=json.loads(data.get('systemPrivileges')),
                propertyPrivileges=PropertyPrivilege(
                            PropertyPrivilegeElement(
                                proPrivilegesData['node'].get('read'),
                                proPrivilegesData['node'].get('write'),
                                proPrivilegesData['node'].get('deny'),
                            ),
                            PropertyPrivilegeElement(
                                proPrivilegesData['edge'].get('read'),
                                proPrivilegesData['edge'].get('write'),
                                proPrivilegesData['edge'].get('deny'),
                            )
                        ),
                policies=ast.literal_eval(data.get('policies')))

            policy_list.append(policy)
        if len(policy_list) == 1:
            return policy_list[0]
        return policy_list

    def asUsers(self) -> List[User]:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
            error = f"DataItem {self.alias} is not Type Table"
            raise ParameterException(error)
        headers = self.entities.getHeaders()
        rows = self.entities.getRows()
        table_rows_dict = convertTableToDict(rows, headers)
        user_list = []
        for data in table_rows_dict:
            proPrivilegesData = json.loads(data.get('propertyPrivileges'))
            user = User(createdTime=datetime.datetime.fromtimestamp(data.get('create')),
                        username=data.get('username'),
                        graphPrivileges=json.loads(data.get('graphPrivileges')),
                        systemPrivileges=json.loads(data.get('systemPrivileges')),
                        propertyPrivileges=PropertyPrivilege(
                            PropertyPrivilegeElement(
                                proPrivilegesData['node'].get('read'),
                                proPrivilegesData['node'].get('write'),
                                proPrivilegesData['node'].get('deny'),
                            ),
                            PropertyPrivilegeElement(
                                proPrivilegesData['edge'].get('read'),
                                proPrivilegesData['edge'].get('write'),
                                proPrivilegesData['edge'].get('deny'),
                            )
                        ),
                        policies=ast.literal_eval(data.get('policies')))

            user_list.append(user)

        if len(user_list) == 1:
            return user_list[0]
        return user_list

    def asStats(self) -> Stats:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
            error = f"DataItem {self.alias} is not Type Table"
            raise ParameterException(error)
        headers = self.entities.getHeaders()
        rows = self.entities.getRows()
        table_rows_dict = convertTableToDict(rows, headers)
        stats_list = []
        for data in table_rows_dict:
            stats = Stats(limitedHdc=data.get('limited_hdc'),
                          limitedShard=data.get('limited_shard'),
                          expiredDate=data.get('expired_date')
                          )

            stats_list.append(stats)

        return stats_list[0]

    def asProcesses(self) -> List[Process]:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
            error = f"DataItem {self.alias} is not Type Table"
            raise ParameterException(error)
        headers = self.entities.getHeaders()
        rows = self.entities.getRows()
        table_rows_dict = convertTableToDict(rows, headers)
        process_list = []
        for data in table_rows_dict:
            process = Process(processId=data.get('process_id'),
                              processQuery=data.get('process_uql'),
                              duration=data.get('duration'),
                              status=data.get('status')
                              )

            process_list.append(process)

        return process_list

    def asTasks(self):
        from ultipa.types import types_response
        _jsonKeys = ['taskJson']
        headers = self.entities.getHeaders()
        rows = self.entities.getRows()
        table_rows_dict = convertTableToDict(rows, headers)
        responseKeyFormat = ResponseKeyFormat(jsonKeys=_jsonKeys)
        if responseKeyFormat:
            table_rows_dict = responseKeyFormat.changeKeyValue(table_rows_dict)
        task_list = []
        for data in table_rows_dict:
            task_info = data.get('taskJson').get('task_info')
            if task_info.get('status_code'):
                task_info["status_code"] = TaskStatusString[task_info.get("TASK_STATUS")]
            if task_info.get('engine_cost'):
                task_info["engine_cost"] = task_info.get("writing_start_time", 0) - task_info.get("start_time", 0)

            data['taskJson']['task_info'] = convertToTask(task_info)
            return_type_get = int(task_info.get('return_type', 0))
            return_type = types_response.Return_Type()
            return_type.is_realtime = True if return_type_get & ALGO_RETURN_TYPE.ALGO_RETURN_REALTIME else False
            return_type.is_visualization = True if return_type_get & ALGO_RETURN_TYPE.ALGO_RETURN_VISUALIZATION else False
            return_type.is_wirte_back = True if return_type_get & ALGO_RETURN_TYPE.ALGO_RETURN_WRITE_BACK else False
            data['taskJson']['task_info'].__setattr__('return_type', return_type)
            task = types_response.Task()
            task.param = data.get('taskJson').get('param')
            task.task_info = data.get('taskJson').get('task_info')
            task.result = data.get('taskJson').get('result')

            task_list.append(task)

        return task_list

    def asProjections(self) -> List[Projection]:
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
            error = f"DataItem {self.alias} is not Type Table"
            raise ParameterException(error)
        return converToProjection(convertTableToDict(self.entities.rows, self.entities.headers))

    def asAny(self) -> any:
        return self.entities

    def asKV(self):
        return self.toDict()

    def asJobs(self):
        if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
            if self.entities is None:
                return []
            return self.entities
        if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
            error = f"DataItem {self.alias} is not Type Table"
            raise ParameterException(error)
        return [converToJob(data) for data in convertTableToDict(self.entities.rows, self.entities.headers)]


class BaseUqlReply:
    def __init__(self, nodes: List[NodeAlias], edges: List[EdgeAlias], tables: List[Table],
                 graphs: List[GraphAlias], attrs: List = None,
                 resultAlias: List = None,
                 explainPlan: ExplainPlan = None,
                 paths: List[PathAlias] = []):
        self.paths = paths
        self.nodes = nodes
        self.edges = edges
        self.tables = tables
        self.attrs = attrs
        self.graphs = graphs
        # self.arrays = arrays
        self.resultAlias = resultAlias
        self.explainPlan = explainPlan


class Statistics(BaseModel):
    '''
        Data class for Statistics
    '''

    def __init__(self, edgeAffected: int = None, nodeAffected: int = None, engineCost: int = None,
                 totalCost: int = None):
        self.edgeAffected = edgeAffected
        self.nodeAffected = nodeAffected
        self.engineCost = engineCost
        self.totalCost = totalCost


class UqlReply(BaseModel):
    datas: List[DataItem]

    def __init__(self, dataBase: BaseUqlReply, aliasMap: dict = None, datas: List = None):
        if aliasMap == None:
            aliasMap = {}
        self._aliasMap = aliasMap
        if datas is None:
            datas = []
        self.datas: List[DataItem] = datas
        self.explainPlan: ExplainPlan = []
        self._dataBase = dataBase

        for data in self._dataBase.paths:
            if self._aliasMap.get(data.alias):
                self._aliasMap[data.alias].data.extend(data.paths)
                continue
            self._aliasMap[data.alias] = DataItem(data.alias, data.paths,
                                                  ResultType.getTypeStr(ResultType.RESULT_TYPE_PATH))

        for data in self._dataBase.graphs:
            if self._aliasMap.get(data.alias):
                self._aliasMap[data.alias].data.extend(data.graph)
                continue
            self._aliasMap[data.alias] = DataItem(data.alias, data.graph,
                                                  ResultType.getTypeStr(ResultType.RESULT_TYPE_GRAPH))

        # for data in self._dataBase.graphs:
        # 	if self._aliasMap.get(data.alias):
        # 		self._aliasMap[data.alias].data.extend(data.graph)
        # 		continue
        # 	self._aliasMap[data.alias] = DataItem(data.alias, data.graph,
        # 										  ResultType.getTypeStr(ResultType.RESULT_TYPE_GRAPH))

        for data in self._dataBase.nodes:
            if self._aliasMap.get(data.alias):
                self._aliasMap[data.alias].data.extend(data.nodes)
                continue
            self._aliasMap[data.alias] = DataItem(data.alias, data.nodes,
                                                  ResultType.getTypeStr(ResultType.RESULT_TYPE_NODE))

        for data in self._dataBase.edges:
            if self._aliasMap.get(data.alias):
                self._aliasMap[data.alias].data.extend(data.edges)
                continue
            self._aliasMap[data.alias] = DataItem(data.alias, data.edges,
                                                  ResultType.getTypeStr(ResultType.RESULT_TYPE_EDGE))

        for data in self._dataBase.attrs:
            if self._aliasMap.get(data.name):
                self._aliasMap[data.name].data.append(data)
                continue
            self._aliasMap[data.name] = DataItem(data.name, data, ResultType.getTypeStr(ResultType.RESULT_TYPE_ATTR))

        for data in self._dataBase.tables:
            if self._aliasMap.get(data.name):
                self._aliasMap[data.name].data.extend(data)
                continue
            self._aliasMap[data.name] = DataItem(data.name, data, ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE))

        for data in self._dataBase.explainPlan.planNodes:
            self.explainPlan.planNodes.append(data)

        for data in self._dataBase.resultAlias:
            if self._aliasMap.get(data.alias):
                self.datas.append(self._aliasMap[data.alias])
        if not self.datas:
            for key in self._aliasMap:
                self.datas.append(self._aliasMap[key])


class ReturnReq:
    def __init__(self, graphSetName: str, uql: str, host: str, retry: Retry, uqlIsExtra: bool):
        self.graph_name = graphSetName
        self.uql = uql
        self.host = host
        self.Retry = retry
        self.uqlIsExtra = uqlIsExtra


class ExportReply:
    def __init__(self, data: List[NodeAlias]):
        self.data = data


class PaserAttrListData:
    def __init__(self, type, nodes: List[Node] = None, edges: List[Edge] = None, paths: List[Path] = None,
                 attrs: UltipaAttr = None):
        self.type = type
        self.nodes = nodes
        self.edges = edges
        self.paths = paths
        self.attrs = attrs


class HDCBuilder:
    def __init__(self, hdcGraphName: str,
                 nodeSchema: [Dict[str, List[str]]],
                 edgeSchema: [Dict[str, List[str]]],
                 hdcServerName: str,
                 syncType: HDCSyncType
                 ):
        self.hdcGraphName = hdcGraphName
        self.nodeSchema = nodeSchema
        self.edgeSchema = edgeSchema
        self.hdcServerName = hdcServerName
        self.syncType = syncType



# -*- coding: utf-8 -*-
# @Time    : 2023/8/1 11:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Schema.py
import json
from typing import List, Dict
from ultipa.structs import DBType
from ultipa.structs.Property import Property
from ultipa.structs.BaseModel import BaseModel


class SchemaStat:
    def __init__(self, type: DBType, name: str, fromSchema: str, toSchema: str,
                 count: int):
        self.type = type
        self.name = name
        self.fromSchema = fromSchema
        self.toSchema = toSchema
        self.count = count

class SchemaStats:
    '''
        Data class for SchemaStats.
    '''

    def __init__(self, stats: List[SchemaStat] = None):
        self.stats = stats


class Schema(BaseModel):
    '''
        Data class for schema.
    '''

    def __init__(self,
                 name: str = None,
                 dbType: DBType = None,
                 description: str = None,
                 properties: List[Property] = None,
                 id: int = None,
                 total: int = 0,
                 stats: List[SchemaStat] = None):
        self.name = name
        self.description = description
        self.properties = properties
        self.dbType = dbType
        self.id = id
        self.total = total
        self.stats = stats

    def getProperty(self, name: str):
        find = list(filter(lambda x: x.get('name') == name, self.properties))
        if find:
            return find[0]
        return None



from enum import Enum


class InsertErrorCode(Enum):


    ID_NOT_MATCH_UUID = 10001
    ID_UUID_NOT_MATCH_SCHEMA = 10002
    FROM_ID_NOT_EXISTED = 10003
    TO_ID_NOT_EXISTED = 10004
    ID_LEN = 10005
    NOT_NULL = 10006
    UNIQUCHECK = 10007
    ID_EMPTY = 10008
    FROM_ID_EMPTY = 10009
    TO_ID_EMPTY = 10010
    DUPLICATE_ID = 10011
    KEY_CONSTRAINT_VIOLATED = 10012
    OK_BUT_ID_EXISTED = 11001
    OTHERS = 19999


class InsertErrorCodeMap:
    codeMap = {
        10001: InsertErrorCode.ID_NOT_MATCH_UUID,
        10002: InsertErrorCode.ID_UUID_NOT_MATCH_SCHEMA,
        10003: InsertErrorCode.FROM_ID_NOT_EXISTED,
        10004: InsertErrorCode.TO_ID_NOT_EXISTED,
        10005: InsertErrorCode.ID_LEN,
        10006: InsertErrorCode.NOT_NULL,
        10007: InsertErrorCode.UNIQUCHECK,
        10008: InsertErrorCode.ID_EMPTY,
        10009: InsertErrorCode.FROM_ID_EMPTY,
        10010: InsertErrorCode.TO_ID_EMPTY,
        10011: InsertErrorCode.DUPLICATE_ID,
        10012: InsertErrorCode.KEY_CONSTRAINT_VIOLATED,
        11001: InsertErrorCode.OK_BUT_ID_EXISTED,
        19999: InsertErrorCode.OTHERS
    }
    @staticmethod
    def getUnsertErrorCode(v):
        return InsertErrorCodeMap.codeMap.get(v)


# -*- coding: utf-8 -*-
# @Time    : 2023/8/4 16:36
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Algo.py
from typing import List

from ultipa.structs.BaseModel import BaseModel


class AlgoParam:
    name: str
    desc: str


class Algo(BaseModel):
    '''
        Data class for algorithm.
    '''

    def __init__(self,
                 name: str = None,
                 type: str = None,
                 writeSupportType:str = None,
                 canRollback: str = None,
                 configContext: str = None,
                 version: str = None,
                 params: List[AlgoParam] = None,
                 description:str = None
                 ):
        self.name = name
        self.type = type
        self.writeSupportType = writeSupportType
        self.canRollback = canRollback
        self.configContext = configContext
        self.version = version
        self.params = params
        self.description = description


class AlgoResultOpt(BaseModel):
    can_realtime: bool
    can_visualization: bool
    can_write_back: bool




from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA_REQUEST, ULTIPA
from ultipa.types.types_response import *
from ultipa.utils import UQLMAKER, CommandList
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.structs.InsertType import InsertType
from ultipa.configuration.InsertRequestConfig import InsertRequestConfig
from ultipa.structs.Node import Node
from ultipa.utils.ufilter.new_ufilter import *

class NodeExtra(BaseExtra):
	'''
	Processing class that defines settings for node related operations.

	'''
	def insertNodes(self, nodes: List[Node], schemaName: str, config:InsertRequestConfig) -> Response:
		'''
		Insert nodes.

		Args:
			nodes: List of nodes to be inserted

			schemaName: The name of the Schema

			config: An object of InsertConfig class

		Returns:
			Response
		'''

		combined_values = []
		for node in nodes:
			node_dict = {}
			if node.id:
				node_dict['_id'] = node.id
			node_dict.update(node.values)
			combined_values.append(node_dict)

		nodes=combined_values
		schemaName='@' + schemaName

		uqlMaker = UQLMAKER(command=CommandList.insert, commonParams=config)
		if config.insertType==InsertType.UPSERT:
			uqlMaker = UQLMAKER(command=CommandList.upsert, commonParams=config)
		if config.insertType==InsertType.OVERWRITE:
			uqlMaker.addParam('overwrite', "", required=False)
		if schemaName:
			uqlMaker.addParam('into', schemaName, required=False)

		uqlMaker.addParam('nodes', nodes)

		if config.silent==False:
			uqlMaker.addParam('as', "nodes")
			uqlMaker.addParam('return', "nodes{*}")

		res = self.uqlSingle(uqlMaker)
		return res




import datetime
import re
import time
from dateutil.parser import parse
from ultipa.utils import errors
from ultipa.utils.errors import ParameterException


class UTC(datetime.tzinfo):
	def __init__(self, offsetHours=0, offsetSeconds=0):
		if offsetSeconds is None:
			self.__offset = 0
		else:
			self.__offset = offsetSeconds
		self.__offsetHour = offsetHours

	def utcoffset(self, dt):
		return datetime.timedelta(seconds=self.__offset, hours=self.__offsetHour)

	def tzname(self, dt):
		return 'UTC+%s dt: %s' % (self.__offsetHour, [dt, id(dt)])

	def dst(self, dt):
		return datetime.timedelta(seconds=self.__offset, hours=self.__offsetHour)

def getTimeZoneSeconds(timeZone):
	import pytz
	try:
		tz = pytz.timezone(timeZone)
		utc_offset = tz.utcoffset(datetime.datetime.utcnow())
		return utc_offset.total_seconds()
	except pytz.exceptions.UnknownTimeZoneError as e:
		raise errors.ParameterException("UnknownTimeZoneError:" + str(e))

def getTimeOffsetSeconds(timeZoneOffset):
	if timeZoneOffset is None:
		return timeZoneOffset
	if isinstance(timeZoneOffset, int):
		return timeZoneOffset
	elif isinstance(timeZoneOffset, float):
		return timeZoneOffset
	elif isinstance(timeZoneOffset, str):
		pattern = re.compile(r"([+-])(\d{2})(\d{2})")
		match = pattern.match(timeZoneOffset)
		if match:
			sign = match.group(1)
			hours = int(match.group(2))
			minutes = int(match.group(3))
			total_offset_minutes = (hours * 60 + minutes) * (-1 if sign == '-' else 1)
			offset = datetime.timedelta(minutes=total_offset_minutes)
			return offset.total_seconds()
		else:
			raise errors.ParameterException("UnknownTimeZoneOffsetError:" + str(timeZoneOffset))
	else:
		raise errors.ParameterException("UnknownTimeZoneOffsetError:" + str(timeZoneOffset))

class UltipaDatetime:
	'''
	Processing class for date and time related operations.
	'''
	year = 0
	month = 0
	day = 0
	hour = 0
	minute = 0
	second = 0
	microsecond = 0

	@classmethod
	def datetimeStr2datetimeInt(self, strDatetime: datetime.datetime):
		'''
		Convert a datatime string into a customized datatime integer.

		Args:
			strDatetime:

		Returns:

		'''

		if isinstance(strDatetime, str):
			try:
				data = datetime.datetime.strptime(strDatetime, "%Y-%m-%d %H:%M:%S.%f%z")
			except:
				try:
					data = datetime.datetime.strptime(strDatetime, "%Y-%m-%d %H:%M:%S.%f")
				except ValueError as e:
					try:
						data = datetime.datetime.strptime(strDatetime, "%Y-%m-%d %H:%M:%S")
					except ValueError as e:
						try:
							data = datetime.datetime.strptime(strDatetime, "%Y-%m-%d")
						except ValueError as e:
							raise ParameterException(e)

		elif isinstance(strDatetime, datetime.datetime):
			data = strDatetime
		else:
			raise ParameterException('strDatetime must str %Y-%m-%d %H:%M:%S.%f or datetime type')
		self.year = data.year
		self.month = data.month
		self.day = data.day
		self.hour = data.hour
		self.minute = data.minute
		self.second = data.second
		self.microsecond = data.microsecond

		if self.year >= 70 and self.year < 100:
			self.year += 1900
		elif self.year < 70:
			self.year += 2000

		datetime_int = 0
		year_month = self.year * 13 + self.month
		datetime_int |= (year_month << 46)
		datetime_int |= (self.day << 41)
		datetime_int |= (self.hour << 36)
		datetime_int |= (self.minute << 30)
		datetime_int |= (self.second << 24)
		datetime_int |= self.microsecond
		return datetime_int

	@classmethod
	def timestampStr2timestampInt(self, strDatetime: str, timeZone, timeZoneOffset=0):
		'''
		Convert strings of datetime, timezone and timezone-offset into the Unix timestamp integer in seconds.

		Args:
			strDatetime:
			timeZone:
			timeZoneOffset:

		Returns:

		'''

		try:
			tzinfo = None
			dt = parse(strDatetime)
			if dt.utcoffset() is not None:
				offset_hours = int(dt.utcoffset().total_seconds() / 3600)
				tzinfo = UTC(offset_hours, timeZoneOffset)

			if timeZone is not None:
				total_seconds = getTimeZoneSeconds(timeZone)
				tzinfo = UTC(0, total_seconds)
			else:
				if timeZoneOffset is not None:
					total_seconds = getTimeOffsetSeconds(timeZoneOffset)
					tzinfo = UTC(0, total_seconds)

			timestamp = datetime.datetime(year=dt.year, month=dt.month, day=dt.day, hour=dt.hour, minute=dt.minute,
										  second=dt.second, tzinfo=tzinfo)
			return int(timestamp.timestamp())
		except ValueError as e:
			raise ParameterException(e)

	@staticmethod
	def datetimeInt2datetimeStr(datetime_int):
		'''
		Convert the customized datetime integer into a datetime string

		Args:
			datetime_int:

		Returns:

		'''
		if datetime_int < 0:
			return ""
		from datetime import datetime
		year_month = ((datetime_int >> 46) & 0x1FFFF)
		year = year_month // 13
		month = year_month % 13
		day = ((datetime_int >> 41) & 0x1F)
		hour = ((datetime_int >> 36) & 0x1F)
		minute = ((datetime_int >> 30) & 0x3F)
		second = ((datetime_int >> 24) & 0x3F)
		microsecond = (datetime_int & 0xFFFFFF)

		def pixString(s, length):
			s = "000000" + str(s)
			return s[len(s) - length:]

		if year == 0:
			return f"{pixString(year, 4)}-{pixString(month, 2)}-{pixString(day, 2)} {pixString(hour, 2)}:{pixString(minute, 2)}:{pixString(second, 2)}.{pixString(microsecond, 2)}"
		if microsecond == 000000:
			ret = datetime(year, month, day, hour, minute, second)
			ret = ret.strftime("%Y-%m-%d %H:%M:%S")
		else:
			ret = datetime(year, month, day, hour, minute, second, microsecond)
			ret = ret.strftime("%Y-%m-%d %H:%M:%S.%f")
		return ret

	@staticmethod
	def timestampInt2timestampStr(datetime_int, timeZone: str = None, timeZoneOffset: int = 0):
		timeStamp = float(datetime_int)
		offset_hours = 0
		if timeZone:
			total_seconds = getTimeZoneSeconds(timeZone)
			offset_hours = int(total_seconds / 3600)
		if timeZoneOffset is None:
			timeZoneOffset = 0
		otherStyleTime = datetime.datetime.fromtimestamp(timeStamp, tz=UTC(offset_hours, timeZoneOffset))
		return str(otherStyleTime)


def wrapper(func):
	'''
	Measures the execution time of a method.

	Args:
		func:

	Returns:

	'''
	
	def inner(*args, **kwargs):
		start_time = time.time()
		res = func(*args, **kwargs)
		end_time = time.time()
		result = end_time - start_time
		print('func %s time is: %.3fs' % (func.__name__, result))
		return res

	return inner


class DateTimestamp(object):
	"""
	A class that realizes the mutual conversion of datetime and timestamp.

	"""

	def __init__(self, date=None):
		if date is None:
			self.timestamp = int(time.time())
			self.datetime = self._toDatetime(self.timestamp)
		else:
			### Judge whether the input date is a timestamp ###
			if isinstance(date, int):
				self.timestamp = date
				self.datetime = self._toDatetime(date)
			else:
				self.timestamp = self._toTimestamp(date)
				self.datetime = date

		if self.timestamp == False:
			self.year = self.month = self.day = self.hour = self.minute = self.second = False
		else:
			self._localtime = time.localtime(self.timestamp) # Parse the tuples from the timestamp
			self.year = self._localtime.tm_year  # Assign year tuple
			self.month = self._localtime.tm_mon  # Assign month tuple
			self.day = self._localtime.tm_mday  # Assign day tuple
			self.hour = self._localtime.tm_hour  # Assign hour tuple
			self.minute = self._localtime.tm_min  # Assign minute tuple
			self.second = self._localtime.tm_sec  # Assign second tuple

	def _toDatetime(self, timestamp):
		"""
		Convert timestamp to datetime

		"""
		try:
			timeStamp = float(timestamp)
			timeArray = time.localtime(timeStamp)
			return time.strftime("%Y-%m-%d %H:%M:%S", timeArray)
		except:
			return False

	def _toTimestamp(self, datetimeString):
		"""
		Convert datetime to timestamp

		"""
		try:
			return int(time.mktime(time.strptime(datetimeString, "%Y-%m-%d %H:%M:%S")))
		except:
			return False

	def __str__(self):
		return self.datetime




def getTimeZoneOffset(requestConfig, defaultConfig):
	'''
	Get timezone offset

	Args:
		requestConfig:

		defaultConfig:

	Returns:

	'''

	timeZone = requestConfig.timeZone
	if timeZone is not None:
		return getTimeZoneSeconds(timeZone)
	timeZoneOffset = requestConfig.timeZoneOffset
	return getTimeOffsetSeconds(timeZoneOffset)



# @Time    : 2023/8/4 10:34
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : QLType.py
from ultipa.proto import ultipa_pb2
class QLType:
    '''
    	    Data class for QL type
    	'''
    UQL = ultipa_pb2.UQL
    GQL = ultipa_pb2.GQL


# -*- coding: utf-8 -*-
# @Time    : 2023/8/1 10:46
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : requestConfig.py
from ultipa.structs.Retry import Retry


class RequestConfig:
	'''
    Configuration class for the interface of any non-insert operation.

    This class provides settings for all the operations other than inserting metadata.
    
    Args:
        - graph (str):Name of the graph to use. If not specified, the graph defined in UltipaConfig.defaultGraph will be used.
        - timeout (int): Request timeout threshold (in second).
        - host (str): Specifies a host in a database cluster to execute the request.
        - thread (int): Number of threads for the request.
        - timeZone (str): Name of the timezone, e.g., Europe/Paris. Defaults to the local timezone if not specified.
        - timeZoneOffset (any): The offset from UTC (in hour), e.g., +2, -3.5. When both timezone and timezoneOffset are provided, timezoneOffset takes precedence.
	'''

	def __init__(self,
				 graph: str = '',
				 timeout: int = None,
				 host: str = None,
				 thread: int = None,
				 timeZone: str = None,
				 timeZoneOffset: int = None):
		self.graph = graph
		self.timeout = timeout
		self.host = host
		self.thread = thread
		self.timeZone = timeZone
		self.timeZoneOffset = timeZoneOffset




from ultipa import DBType, UltipaPropertyType
from ultipa.types.types import ErrorCode, Schemas
from ultipa.utils.convert import  convertToSchemas
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA
from ultipa.types.types_response import *
from ultipa.utils import UQLMAKER, CommandList
from typing import Tuple, List, Union
from ultipa.operate.property_extra import PropertyExtra
from ultipa.utils.format import FormatType
from ultipa.utils.noneCheck import checkNone

BOOL_KEYS = ["index", "lte"]
REPLACE_KEYS = {
	"name": "schemaName",
	"type": "propertyType",
}


class SchemaExtra(BaseExtra):
	'''
		Prcessing class that defines settings for schema related operations.
	'''

	def createSchema(self, schema: Schema, isCreateProperties: bool = False,
					 config: RequestConfig = RequestConfig()) -> Response:
		'''
		Create a schema.

		Args:
			schema: The schema to be created; the attributes name and dbType are mandatory, properties and description are optional.

			isCreateProperties: Whether to create properties associated with the schema, the default is False.

			config: An object of RequestConfig class

		Returns:
			Response

		'''

		if checkNone(schema.name) or checkNone(schema.dbType):
			return Response(status=Status(code = ULTIPA.ErrorCode.PARAM_ERROR ,message='The name and dbType of the schema cannot be None'))

		command = schema.dbType == DBType.DBNODE and CommandList.createNodeSchema or CommandList.createEdgeSchema
		commandP = [f"`{schema.name}`"]
		if schema.description:
			commandP.append(schema.description)
		uqlMaker = UQLMAKER(command=command, commonParams=config)
		uqlMaker.setCommandParams(commandP=commandP)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		if res.status.code != ErrorCode.SUCCESS:
			return res
		if isCreateProperties:
			if schema.properties:
				for prop in schema.properties:
					try:
						res1 = PropertyExtra.createProperty(self, dbType=schema.dbType, schemaName=schema.name,
															property=prop, config=config)
					except Exception as e:
						print("An error occurred while creating property:", prop.name, "Error:", e)
		return res


	def showSchema(self,
				   config: RequestConfig = RequestConfig()) -> List[Schema]:
		'''
		Show schema(s).

		Args:

			config: An object of RequestConfig class

		Returns:
			List[Schema]

		'''

		command = CommandList.showSchema
		commandP = ''

		uqlMaker = UQLMAKER(command=command, commandP=commandP, commonParams=config)
		res = self.uqlSingle(uqlMaker)
		if res.status.code == ULTIPA.ErrorCode.SUCCESS and res.items:
			schemas = convertToSchemas(res)
			return schemas.schema
		return res

	def showNodeSchema(self, config: RequestConfig = RequestConfig()) -> List[Schema]:
		'''
        Show Nodeschema(s).

        Args:
            config: An object of RequestConfig class

        Returns:
            List[Schema]

        '''

		command = CommandList.showNodeSchema
		commandP = ''

		uqlMaker = UQLMAKER(command=command, commandP=commandP, commonParams=config)
		res = self.uqlSingle(uqlMaker)
		if res.status.code == ULTIPA.ErrorCode.SUCCESS and res.items:
			schemas = convertToSchemas(res)
			return schemas.schema
		return res


	def showEdgeSchema(self, config: RequestConfig = RequestConfig()) -> List[Schema]:
		'''
        Show Edgeschema(s).

        Args:
            config: An object of RequestConfig class

        Returns:
            List[Schema]

        '''

		command = CommandList.showEdgeSchema
		commandP = ''

		uqlMaker = UQLMAKER(command=command, commandP=commandP, commonParams=config)
		res = self.uqlSingle(uqlMaker)
		if res.status.code == ULTIPA.ErrorCode.SUCCESS and res.items:
			schemas = convertToSchemas(res)
			return schemas.schema
		return res

	def alterSchema(self, originalSchema: Schema, newSchema: Schema,
					config: RequestConfig = RequestConfig()) -> Response:
		'''
		Alter schema.

		Args:
			originalSchema: The schema to be altered; the attributes name and dbType are mandatory

			newSchema: Schema: A Schema object used to set the new name and/or description for the schema.

			config: An object of RequestConfig class

		Returns:
			Response
		'''
		if checkNone(originalSchema.name) or checkNone(originalSchema.dbType):
			return Response(status = Status(code=ErrorCode.PARAM_ERROR,message='The name and dbType of originalSchema cannot be None'))
		if newSchema.dbType is not None and newSchema.dbType != originalSchema.dbType:
			return Response(status = Status(code=ErrorCode.PARAM_ERROR,message='The new dbType must be consistent with the old dbType'))
		command = originalSchema.dbType == DBType.DBNODE and CommandList.alterNodeSchema or CommandList.alterEdgeSchema
		commandP = "@`%s`" % (originalSchema.name)
		update_dict = {}
		if newSchema.name:
			update_dict.setdefault('name', newSchema.name)
		if newSchema.description:
			update_dict.update({'description': newSchema.description})
		uqlMaker = UQLMAKER(command=command, commandP=commandP, commonParams=config)
		uqlMaker.addParam("set", update_dict)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def dropSchema(self, schema: Schema,
				   config: RequestConfig = RequestConfig()) -> Response:
		'''
		Drop schema.

		Args:
			schema: The schema to be dropped; the attributes name and dbType are mandatory.

			config: An object of RequestConfig class

		Returns:
			Response

		'''
		if checkNone(schema.name) or checkNone(schema.dbType):
			return Response(status = Status(code=ErrorCode.PARAM_ERROR,message='The name and dbType of schema cannot be None'))
		command = schema.dbType == DBType.DBNODE and CommandList.dropNodeSchema or CommandList.dropEdgeSchema
		commandP = "@`%s`" % (schema.name)

		uqlMaker = UQLMAKER(command=command, commandP=commandP, commonParams=config)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def getSchema(self, schemaName: str, dbType: DBType,
				  config: RequestConfig = RequestConfig()) -> Schema:
		'''
		Acquire a designated Schema.

		Args:
			schemaName: The name of Schema

			dbType: The DBType of data (DBNODE or DBEDGE)

			config: An object of RequestConfig class

		Returns:
			Schema
		'''

		if FormatType.checkNone(schemaName) or FormatType.checkNone(dbType):
			message = 'The schemaName and dbType cannot None' if FormatType.checkNone(schemaName) and FormatType.checkNone(dbType) else 'The dbType cannot None' if FormatType.checkNone(dbType) else 'The schemaName cannot None'

			return Response(status=Status(code=ErrorCode.PARAM_ERROR, message=message))
		if dbType != None:
			if dbType == DBType.DBNODE:
				command = CommandList.showNodeSchema
			elif dbType == DBType.DBEDGE:
				command = CommandList.showEdgeSchema
			elif dbType == DBType.DBGLOBAL:
				return Response(status=Status(code=ErrorCode.PARAM_ERROR, message='The dbType cannot be DBType.DBGLOBAL'))

			if schemaName:
				commandP = "@`%s`" % (schemaName)

		uqlMaker = UQLMAKER(command, commandP=commandP, commonParams=config)

		res = self.uqlSingle(uqlMaker=uqlMaker)
		if res.status.code == ULTIPA.ErrorCode.SUCCESS and res.items:
			schemas = convertToSchemas(res)
			return schemas.schema[0]
		return res.items

	def getNodeSchema(self, schemaName: str,
					  config: RequestConfig = RequestConfig()) -> Schema:
		'''
        Acquire a designated Node Schema.

        Args:
            schemaName: The name of Schema

            config: An object of RequestConfig class

        Returns:
            Schema
        '''

		return self.getSchema(dbType=DBType.DBNODE,schemaName = schemaName,config=config)

	def getEdgeSchema(self, schemaName: str,
					  config: RequestConfig = RequestConfig()) -> Schema:
		'''
        Acquire a designated Edge Schema.

        Args:
            schemaName: The name of Schema

            config: An object of RequestConfig class

        Returns:
            Schema
        '''

		return self.getSchema(dbType=DBType.DBEDGE,schemaName = schemaName,config=config)

	def createSchemaIfNotExist(self, schema: Schema,isCreateProperties: bool,
							   config: RequestConfig = RequestConfig()) -> ResponseWithExistCheck:
		'''
        Create a schema if schema does not exist.

        Args:
            schema: The schema to be created; the attributes name and dbType are mandatory, properties and description are optional.

            config: An object of RequestConfig class

        Returns:
            ResponseWithExistCheck

        '''
		if checkNone(schema.name) or checkNone(schema.dbType):
			return Response(status = Status(code=ErrorCode.PARAM_ERROR,message='The name and dbType of originalSchema cannot be None'))
		check = self.getSchema(schemaName=schema.name, dbType=schema.dbType, config=config)
		if check is not None:
			return ResponseWithExistCheck(exist=True,response=Response())

		elif check is None:
			res = self.createSchema(schema=schema,isCreateProperties=isCreateProperties, config=config)
			return ResponseWithExistCheck(exist=False,response=res)

	def getProperty(self, dbType: DBType, propertyName: str, schemaName: str,
					config: RequestConfig = RequestConfig()) -> Property:
		'''
        Get a property.

        Args:
            dbType: The DBType of data (DBNODE or DBEDGE)

            schemaName: The name of schema

            propertyName: The name of the Property

            config: An object of RequestConfig class

        Returns:
            Property

        '''
		if checkNone(dbType) or checkNone(propertyName) or checkNone(schemaName):
			message ='The dbType cannot be None' if checkNone(dbType) else 'The dbType cannot be DBType.DBGLOBAL' if dbType == DBType.DBGLOBAL else 'The propertyName cannot be None' if checkNone(propertyName) else 'The schemaName cannot be None'
			return Response(status=Status(code = ErrorCode.PARAM_ERROR,message=message))
		res = self.getSchema(schemaName=schemaName, dbType=dbType, config=config)
		if res and not isinstance(res,Response) :
			properties = res.properties
			for property in properties:
				if property.name == propertyName:
					property.type = property.getPropertyTypeByString(property.type)
					return property
		return None

	def getNodeProperty(self, propertyName: str, schemaName: str,
						config: RequestConfig = RequestConfig()) -> Property:
		'''
        Get a Node property.

        Args:
            schemaName: The name of schema

            propertyName: The name of the Property

            config: An object of RequestConfig class

        Returns:
            Property

        '''
		res = self.getProperty(propertyName=propertyName, schemaName=schemaName, dbType=DBType.DBNODE,
							   config=config)
		return res

	def getEdgeProperty(self, propertyName: str, schemaName: str,
						config: RequestConfig = RequestConfig()) -> Property:
		'''
        Get an Edge property.

        Args:
            schemaName: The name of schema

            propertyName: The name of the Property

            config: An object of RequestConfig class

        Returns:
            Property

        '''
		res = self.getProperty(propertyName=propertyName, schemaName=schemaName, dbType=DBType.DBEDGE,
							   config=config)
		return res

	def createPropertyIfNotExist(self, dbType: DBType, schemaName: str, property: Property,
								 config: RequestConfig = RequestConfig()) -> ResponseWithExistCheck:
		'''
		Create a schema if schema does not exist.

		Args:
			dbType: Type of the property (node or edge).

			schemaName: An object of Schema class

			property: The property to be created; the attributes name and type (and subType if the type is SET or LIST) are mandatory, encrypt and description are optional.

			config: An object of RequestConfig class

		Returns:
			ResponseWithExistCheck

		'''
		if (property.type == UltipaPropertyType.LIST or property.type == UltipaPropertyType.SET) and checkNone(property.subType):
			return Response(status=Status(code=ErrorCode.PARAM_ERROR,message='The subType of property cannot be None'))

		res = self.getProperty(dbType=dbType, propertyName=property.name, schemaName=schemaName,
							   config=config)
		if res != None:
			return ResponseWithExistCheck(exist=True,response=Response())
		if res == None:
			newprop = PropertyExtra.createProperty(self, dbType=dbType, schemaName=schemaName, property=property,
												   config=config)
			return ResponseWithExistCheck(exist=False,response=newprop)


# -*- coding: utf-8 -*-
# @Time    : 2023/8/4 10:29
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : BaseModel.py
import json
from datetime import datetime
from enum import Enum
from typing import Dict


class BaseModel:
	'''
		Processing class that defines settings for returned data related operations.
	'''
	def toJSON(self, pretty=False):
		def custom_default(o):
			if isinstance(o, Enum):
				return o.name
			elif isinstance(o,datetime) :
				return o.strftime("%Y-%m-%d %H:%M:%S")
			else:
				return o.__dict__
		try:
			if pretty:
				return json.dumps(self, default=lambda o: o.__dict__,
								  sort_keys=True, indent=4, ensure_ascii=False)
			else:

				return json.dumps(self, default=lambda o: custom_default(o) , sort_keys=False, ensure_ascii=False)

		except Exception as e:
			return self

	def toDict(self):
		return json.loads(self.toJSON())

	def __str__(self):
		return str(self.__dict__)



from typing import List, Optional, Union

from ultipa.structs.Process import Process
from ultipa.types.types_response import Task

from ultipa.types.types import TaskStatus

from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA
from ultipa.types.types_response import *
from ultipa.utils import UQLMAKER, CommandList
from ultipa.structs.Job import Job
import json

from ultipa.utils.convert import convertToAnyObject, convertToTop, convertTableToDict, convertToLicense, converToJob
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from ultipa.structs.Top import Top


class ALGO_RETURN_TYPE:
    ALGO_RETURN_REALTIME = 1
    ALGO_RETURN_WRITE_BACK = 2
    ALGO_RETURN_VISUALIZATION = 4


class TaskExtra(BaseExtra):
    '''
	Processing class that defines settings for task and process related operations.
	'''

    def top(self,
            config: RequestConfig = RequestConfig()) -> List[Process]:
        '''
        Top real-time processes.

        Args:
            config: An object of RequestConfig class

        Returns:
            List[Process]

        '''
        uqlMaker = UQLMAKER(command=CommandList.top, commonParams=config)
        res = self.UqlListSimple(uqlMaker)
        if hasattr(res,'data'):
            if len(res.data) > 0:
                res.data = convertToTop(res)

        else:
            res.data = []
        return res.data

    def kill(self, id: str,
             config: RequestConfig = RequestConfig()) -> Response:
        '''
        Kill real-time processes.

        Args:
            id: ID of the process to kill; set to * to kill all processes.

            config: An object of RequestConfig class

        Returns:
            Response

        '''
        commonP = id
        uqlMaker = UQLMAKER(command=CommandList.kill, commonParams=config)
        uqlMaker.setCommandParams(commonP)
        res = self.uqlSingle(uqlMaker)
        return res

    def showJob(self, id: str = None,
                config: RequestConfig = RequestConfig()) -> List[Job]:
        '''
        Show back-end job.

        Args:
            id: ID of the job.

            config: An object of RequestConfig class

        Returns:
            List[Job]

        '''

        _jsonKeys = ['taskJson']
        uqlMaker = UQLMAKER(command=CommandList.showJob, commonParams=config)
        if id:
            commonP = int(id)
            uqlMaker.setCommandParams(commandP=commonP)
        res = self.UqlListSimple(uqlMaker=uqlMaker, responseKeyFormat=ResponseKeyFormat(jsonKeys=_jsonKeys))
        newDatas = []
        if res.data :
            for obj in res.data:
                obj = obj.__dict__
                job  = converToJob(obj)
                newDatas.append(job)
            res.data = newDatas
        return res.data

    def clearJob(self, id: str,
                 config: RequestConfig = RequestConfig()) -> Response:
        '''
        Clear back-end job.

        Args:
            id: ID of the job.

            config: An object of RequestConfig class

        Returns:
            Response

        '''

        uqlMaker = UQLMAKER(command=CommandList.clearJob, commonParams=config)

        commonP = id
        uqlMaker.setCommandParams(commandP=commonP)

        res = self.uqlSingle(uqlMaker)

        return res

    def stopJob(self, id: str,
                 config: RequestConfig = RequestConfig()) -> Response:
        '''
        Stop back-end job.

        Args:
            id: The ID of back-end job

            config: An object of RequestConfig class

        Returns:
            Response

        '''
        uqlMaker = UQLMAKER(command=CommandList.stopJob, commonParams=config)
        commonP = id
        uqlMaker.setCommandParams(commandP=commonP)
        res = self.uqlSingle(uqlMaker)
        if res.status.code != ULTIPA.ErrorCode.SUCCESS:
            return  Response(status=res.status,statistics=res.statistics)
        return res



from typing import List


class PropertyPrivilegeElement:
    def __init__(self, read: List[List[str]] = None, write: List[List[str]] = None, deny: List[List[str]] = None):
        self.read = read
        self.write = write
        self.deny = deny

    def to_dict(self):
        self.__dict__.items()
        return {k:v for k,v in self.__dict__.items() if v is not None }

class PropertyPrivilege:
    def __init__(self, node: PropertyPrivilegeElement = None, edge: PropertyPrivilegeElement = None):
        self.node = node
        self.edge = edge

    def to_dict(self):
        result = {k:v for k,v in self.__dict__.items() if v is not None }.copy()
        for key, value in result.items():
            if hasattr(value, 'to_dict'):
                result[key] = value.to_dict()
        return result



# -*- coding: utf-8 -*-
# @Time    : 2024/7/5 18:14
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : InsertRequestConfig.py
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.structs.InsertType import InsertType


class InsertRequestConfig(RequestConfig):
	'''
    Configuration class for the interface of insert operation.

    This class provides settings for inserting metadata.
    
    Args:
        - insertType (InsertType): The insertion mode. Supports NORMAL, UPSERT, and OVERWRITE.
        - graph (str):Name of the graph to use. If not specified, the graph defined in UltipaConfig.defaultGraph will be used.
        - timeout (int): Request timeout threshold (in second).
        - host (str): Specifies a host in a database cluster to execute the request.
        - timeZone (str): The string of timezone in standard format
        - timeZoneOffset (any): 1, the number of seconds; 2, a 5-character string such as +0700, -0430
        - silent (bool):Whether to return the _id or _uuid of the operated nodes or edges. Sets to Ture to not return, and False to return.
	'''
	def __init__(self,
				 insertType: InsertType = InsertType.NORMAL,
				 graph: str = '',
				 timeout: int = None,
				 host: str = None,
				 timeZone:str =None,
				 thread:int = None,
				 timeZoneOffset:int =None,
				 silent :bool = True):
		super().__init__(graph, timeout, host,thread, timeZone=timeZone,
						 timeZoneOffset=timeZoneOffset)
		self.insertType = insertType

		self.silent = silent




# -*- coding: utf-8 -*-
# @Time    : 2023/8/4 16:39
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : ResultType.py
from enum import Enum

from ultipa.proto import ultipa_pb2
from ultipa.structs.PropertyType import UltipaPropertyType


class ResultType(Enum):
	'''
		Data class for returned result type.
	'''
	RESULT_TYPE_UNSET = ultipa_pb2.RESULT_TYPE_UNSET
	RESULT_TYPE_PATH = ultipa_pb2.RESULT_TYPE_PATH
	RESULT_TYPE_NODE = ultipa_pb2.RESULT_TYPE_NODE
	RESULT_TYPE_EDGE = ultipa_pb2.RESULT_TYPE_EDGE
	RESULT_TYPE_ATTR = ultipa_pb2.RESULT_TYPE_ATTR
	# RESULT_TYPE_GRAPH = ultipa_pb2.RESULT_TYPE_GRAPH
	# RESULT_TYPE_ARRAY = ultipa_pb2.RESULT_TYPE_ARRAY
	RESULT_TYPE_TABLE = ultipa_pb2.RESULT_TYPE_TABLE
	RESULT_TYPE_GRAPH = 6
	RESULT_TYPE_ExplainPlan = "ExplainPlan"

	@staticmethod
	def getTypeStr(type):
		if type == ResultType.RESULT_TYPE_PATH:
			return ResultType.RESULT_TYPE_PATH
		elif type == ResultType.RESULT_TYPE_NODE:
			return ResultType.RESULT_TYPE_NODE
		elif type == ResultType.RESULT_TYPE_EDGE:
			return ResultType.RESULT_TYPE_EDGE
		elif type == ResultType.RESULT_TYPE_ATTR:
			return ResultType.RESULT_TYPE_ATTR
		elif type == UltipaPropertyType.LIST.value:
			return ResultType.RESULT_TYPE_ATTR
		elif type == ResultType.RESULT_TYPE_TABLE:
			return ResultType.RESULT_TYPE_TABLE
		elif type == ResultType.RESULT_TYPE_UNSET:
			return ResultType.RESULT_TYPE_UNSET
		# elif type == ResultType.RESULT_TYPE_GRAPH:
		# 	return "GRAPH"
		elif type == ResultType.RESULT_TYPE_ExplainPlan:
			return ResultType.RESULT_TYPE_ExplainPlan
		else:
			return type

	@staticmethod
	def getTypeen(type):
		if type == ResultType.RESULT_TYPE_PATH.value:
			return ResultType.RESULT_TYPE_PATH
		elif type == ResultType.RESULT_TYPE_NODE.value:
			return ResultType.RESULT_TYPE_NODE
		elif type == ResultType.RESULT_TYPE_EDGE.value:
			return ResultType.RESULT_TYPE_EDGE
		elif type == ResultType.RESULT_TYPE_ATTR.value:
			return ResultType.RESULT_TYPE_ATTR
		elif type == UltipaPropertyType.LIST.value:
			return ResultType.RESULT_TYPE_ATTR
		elif type == ResultType.RESULT_TYPE_TABLE.value:
			return ResultType.RESULT_TYPE_TABLE
		elif type == ResultType.RESULT_TYPE_UNSET.value:
			return ResultType.RESULT_TYPE_UNSET
		# elif type == ResultType.RESULT_TYPE_GRAPH:
		# 	return "GRAPH"
		elif type == ResultType.RESULT_TYPE_ExplainPlan.value:
			return ResultType.RESULT_TYPE_ExplainPlan
		else:
			return type


from ultipa import UpdateEdge
from ultipa.operate.base_extra import BaseExtra
from ultipa.types.types_response import *

from ultipa.types import  ULTIPA
from ultipa.utils import UQLMAKER, CommandList
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.structs.InsertType import InsertType
from ultipa.configuration.InsertRequestConfig import InsertRequestConfig
from ultipa.structs.Edge import Edge
from ultipa.utils.noneCheck import checkNone
from ultipa.utils.ufilter.new_ufilter import *


class EdgeExtra(BaseExtra):
    '''
    Processing class that defines settings for edge related operations.

    '''
    def insertEdges(self, edges: List[Edge], schemaName: str,
                    config: InsertRequestConfig) -> Response:
        '''
        Insert edges.

        Args:
            edges: The list of edges to be inserted; the attributes fromId and toId of each Edge are mandatory, uuid, fromUuid, and toUuid cannot be set.

            schemaName: Name of the schema

            config: An object of InsertRequestConfig class

        Returns:
            Response
        '''
        combined_values = []  # to combine values and id for insertion
        for edge in edges:
            if checkNone(edge.fromId) or checkNone(edge.toId):
                return Response(status=Status(code = ErrorCode.PARAM_ERROR,message='The fromId and toId of edge cannot be None'))
            edge_dict = {}
            if edge.fromId:
                edge_dict["_from"] = edge.fromId
            if edge.toId:
                edge_dict["_to"] = edge.toId

            if edge.fromUuid:
                edge_dict["_from_uuid"] = edge.fromUuid
            if edge.toUuid:
                edge_dict["_to_uuid"] = edge.toUuid

            if edge.uuid:
                edge_dict["_uuid"] = edge.uuid
            edge_dict.update(edge.values)
            combined_values.append(edge_dict)
        edges = combined_values
        schemaName = '@' + schemaName

        uqlMaker = UQLMAKER(command=CommandList.insert, commonParams=config)
        if config.insertType == InsertType.UPSERT:
            uqlMaker = UQLMAKER(command=CommandList.upsert, commonParams=config)
        if config.insertType == InsertType.OVERWRITE:
            uqlMaker.addParam('overwrite', "", required=False)
        if schemaName:
            uqlMaker.addParam('into', schemaName, required=False)
        uqlMaker.addParam('edges', edges)

        if config.silent == False:
            uqlMaker.addParam('as', "edges")
            uqlMaker.addParam('return', "edges{*}")
        res = self.uqlSingle(uqlMaker)
        return res


# -*- coding: utf-8 -*-
# @Time    : 2024/05/14 11:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Index.py
from typing import List, Dict

from ultipa.structs import DBType
from ultipa.structs.BaseModel import BaseModel


class Job(BaseModel):
    '''
        Data class for Index.
    '''

    def __init__(self, id: str = None,
                 graphName: str = None,
                 type: str = None,
                 query: str = None,
                 status: str = None,
                 errMsg: str = None,
                 result: Dict = {},
                 startTime: str = None,
                 endTime: str = None,
                 progress: str = None):
        self.id = id
        self.graphName = graphName
        self.query = query
        self.type = type
        self.status = status
        self.errMsg = errMsg
        self.result = result
        self.startTime = startTime
        self.endTime = endTime
        self.progress = progress



# -*- coding: utf-8 -*-
# @Time    : 2024/05/17 10:56
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Stats.py
from ultipa.structs.BaseModel import BaseModel


class Stats(BaseModel):
	'''
	    Data class for Statistics.
	'''

	def __init__(self,limitedShard:str = '',limitedHdc:str = '',expiredDate:str = '',
				 licenseUUId: str = '', company: str = '',department: str = '',limitedUser: str = '',limitedGraph: str = '',limitedNode: str = '',
				 limitedEdge:str = ''):
		# self.licenseUUId = licenseUUId
		# self.company = company
		# self.department = department
		# self.limitedUser = limitedUser
		# self.limitedGraph = limitedGraph
		# self.limitedNode = limitedNode
		# self.limitedEdge = limitedEdge
		self.limitedShard = limitedShard
		self.limitedHdc = limitedHdc
		self.expiredDate = expiredDate
	
	
	
	
	
	
	


# -*- coding: utf-8 -*-
# @Time    : 2023/7/18 09:21
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : hostManager.py
import random
from typing import List

from ultipa.connection.clientInfo import GrpcClientInfo
from ultipa.connection.clientType import ClientType
from ultipa.connection.uqlHelper import UQLHelper
from ultipa.types import ULTIPA


class HostManager:
	'''
		Client manager class for managing Ultipa connection on behalf of client end.

	'''
	username: str
	password: str
	crt: str
	graphSetName: str
	leaderHost: str
	followersPeerInfos: List[ULTIPA.RaftPeerInfo] = None
	leaderInfos: ULTIPA.RaftPeerInfo = None
	leaderClientInfo: GrpcClientInfo = None
	algoClientInfos: List[GrpcClientInfo] = []
	defaultClientInfo: GrpcClientInfo = None
	otherFollowerClientInfos: List[GrpcClientInfo] = []
	otherUnsetFollowerClientInfos: List[GrpcClientInfo] = []
	nullClientInfos: GrpcClientInfo = None
	raftReady: bool = False

	def __init__(self, graphSetName: str, host: str, username: str, password: str, crt: str, maxRecvSize: int):
		self.graphSetName = graphSetName
		self.username = username
		self.password = password
		self.maxRecvSize = maxRecvSize
		self.leaderHost = host
		self.crt = crt
		self.defaultClientInfo = self.__createClientInfo(host)
		self.nullClientInfos = self.__createClientInfo('0.0.0.0')

	def __createClientInfo(self, host: str):

		if self.defaultClientInfo and self.defaultClientInfo.host == host:
			return self.defaultClientInfo

		clientInfo = GrpcClientInfo(host=str(random.choices(host.split(','))[0]), username=self.username, password=self.password, crt=self.crt,
									maxRecvSize=self.maxRecvSize)
		return clientInfo

	def chooseClientInfo(self, clientType: int, uql: str,  useHost: str):
		'''
		Three rules: Send to a random node for load balancing.
		Select the connection object.
		:param clientType: the type of connection
		:return:
		'''
		if useHost:
			for clientInfo in self.getAllClientInfos():
				if useHost == clientInfo.host:
					return clientInfo
			return self.__createClientInfo(useHost)



		clientType = clientType or ClientType.Default
		if uql:
			if UQLHelper.uqlIsExecTask(uql):
				clientType = ClientType.Algo
			elif UQLHelper.uqlIsWrite(uql):
				clientType = ClientType.Update
			elif UQLHelper.uqlIsForceUseMaster(uql):
				clientType = ClientType.Leader
		if clientType == ClientType.Algo:
			if not self.algoClientInfos:
				return self.nullClientInfos
			return random.choice(self.algoClientInfos)
		if clientType == ClientType.Update or clientType == ClientType.Leader:
			return self.leaderClientInfo or self.defaultClientInfo

		return random.choice(list(self.getAllClientInfos(ignoreAlgo=True, needUnset=False)))

	def getAllHosts(self):
		hosts = [self.leaderHost]
		if self.followersPeerInfos:
			for PeerHost in self.followersPeerInfos:
				hosts.append(PeerHost.host)
		return hosts

	def getAllClientInfos(self, ignoreAlgo: bool = False, needUnset: bool = True):
		all = [self.defaultClientInfo]
		if self.leaderClientInfo and self.leaderClientInfo.host != self.defaultClientInfo.host:
			all.append(self.leaderClientInfo)
		if self.algoClientInfos and not ignoreAlgo:
			all.extend(self.algoClientInfos)
		if self.otherFollowerClientInfos:
			all.extend(self.otherFollowerClientInfos)
		if self.otherUnsetFollowerClientInfos and needUnset:
			all.extend(self.otherUnsetFollowerClientInfos)
		return all

	def setClients(self, leaderHost: str, followersPeerInfos: List[ULTIPA.RaftPeerInfo],
				   leaderInfos: ULTIPA.RaftPeerInfo):

		self.leaderHost = leaderHost
		self.leaderClientInfo = self.__createClientInfo(leaderHost)
		self.followersPeerInfos = followersPeerInfos
		self.leaderInfos = leaderInfos
		self.otherFollowerClientInfos = []
		self.otherUnsetFollowerClientInfos = []
		self.algoClientInfos = []

		if followersPeerInfos and len(followersPeerInfos) > 0:
			for peerInfo in followersPeerInfos:

				if peerInfo.isAlgoExecutable:
					self.algoClientInfos.append(self.__createClientInfo(peerInfo.host))

				if peerInfo.isFollowerReadable:
					self.otherFollowerClientInfos.append(self.__createClientInfo(peerInfo.host))

				if peerInfo.isUnset:
					self.otherUnsetFollowerClientInfos.append(self.__createClientInfo(peerInfo.host))
		else:
			self.algoClientInfos = [self.leaderClientInfo]



from ultipa import ParameterException
from ultipa.configuration.UltipaConfig import UltipaConfig
from ultipa.operate.download_extra import DownloadExtra
from ultipa.operate.edge_extra import EdgeExtra
from ultipa.operate.export_extra import ExportExtra
from ultipa.operate.index_extra import IndexExtra
from ultipa.operate.lte_ufe_extra import LteUfeExtra
from ultipa.operate.node_extra import NodeExtra
from ultipa.operate.policy_extra import PolicyExtra
from ultipa.operate.task_extra import TaskExtra
from ultipa.operate.truncate_extra import TruncateExtra
from ultipa.operate.user_extra import UserExtra
from ultipa.operate.algo_extra import AlgoExtra
from ultipa.operate.hdc_extra import HDCExtra
from ultipa.types.types import ErrorCode
from ultipa.operate.graph_extra import GraphExtra
from ultipa.operate.property_extra import PropertyExtra
from ultipa.operate.schema_extra import SchemaExtra

class Connection(DownloadExtra, UserExtra, NodeExtra, EdgeExtra, LteUfeExtra, IndexExtra, PolicyExtra,
				 TaskExtra, ExportExtra,PropertyExtra, TruncateExtra,AlgoExtra,HDCExtra,GraphExtra,SchemaExtra):
	'''
		The class of Ultipa connection.

	'''

	def RunHeartBeat(self, time: int):
		self.keepConnectionAlive(time)

	def StopHeartBeat(self):
		self.stopConnectionAlive()

	@staticmethod
	def NewConnection(defaultConfig: UltipaConfig = UltipaConfig()):
		conn = None
		if not defaultConfig.hosts:
			raise ParameterException(err="hosts is a required parameter")
		if not defaultConfig.username:
			raise ParameterException(err="username is a required parameter")
		if not defaultConfig.password:
			raise ParameterException(err="password is a required parameter")
		for host in defaultConfig.hosts:
			conn = Connection(host=host, defaultConfig=defaultConfig, crt=defaultConfig.crt)
			testRes = conn.test()
			if testRes.status.code == ErrorCode.SUCCESS:
				if defaultConfig.heartBeat > 0:
					conn.RunHeartBeat(defaultConfig.heartBeat)
				return conn
		return conn

	@staticmethod
	def GetConnection(defaultConfig: UltipaConfig = UltipaConfig()):
		return Connection.NewConnection(defaultConfig)

	def __del__(self):
		self.StopHeartBeat()



from ultipa import ExportRequest
from ultipa.operate.base_extra import BaseExtra
from ultipa.configuration.RequestConfig import RequestConfig
from typing import Callable,List
from ultipa.structs.Node import Node
from ultipa.structs.Edge import Edge
class ExportExtra(BaseExtra):
	'''
		Processing class that defines settings for data exporting operation.
	'''

	def export(self, exportRequest: ExportRequest,cb: Callable[[List[Node], List[Edge]], None],
			   config: RequestConfig = RequestConfig()) :
		'''
		Export data.

		Args:
			exportRequest: An object of ExportRequest class

			cb: callback function for receiving data

			config: An object of RequestConfig class

		Returns:
			Stream
		'''
		self.exportData(exportRequest,cb,config)
		return



# -*- coding: utf-8 -*-
# @Time    : 2023/7/18 11:53
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : clientType.py
class ClientType:
	'''
		This class defines the types of client.

	'''
	Default = 0
	Algo = 1
	Update = 2
	Leader = 3



from typing import List, Dict
from ultipa.printer.prettyPrint import PrettyPrint, PrettyTable
from ultipa.structs.GraphSet import GraphSet
from ultipa.structs.InsertErrorCode import InsertErrorCode
from ultipa.structs.Property import Property
from ultipa.structs.Schema import Schema
from ultipa.structs.User import User
from ultipa.structs.Policy import Policy
from ultipa.structs.Privilege import Privilege
from ultipa.structs.Index import Index
from ultipa.structs.Stats import Stats
from ultipa.structs.Top import Top
from ultipa.structs.Algo import Algo
from ultipa.types.types import BaseModel, Node, Edge, DataItem, Status, \
    Statistics, ReturnReq, BaseUqlReply, ResultType, Alias, ExplainPlan,ErrorCode
from ultipa.utils.convert import Any
from typing import Callable


class PropertyTable(BaseModel):
    name: str
    data: List[Property]

    def __init__(self, name, data):
        self.name = name
        self.data = data


class UltipaResponse(BaseModel):
    def __init__(self, status: Status = None, data: BaseUqlReply = None,
                 req: ReturnReq = None, statistics: Statistics = None, aliases: List[Alias] = None):
        self.status = status
        self.data = data
        self.statistics = statistics
        self.req = req
        self.aliases = aliases

    def Print(self):
        pretty = PrettyPrint()
        pretty.prettyStatus(self.status)
        if self.status.code != ErrorCode.SUCCESS:
            return
        if self.statistics:
            pretty.prettyStatistics(self.statistics)

        if isinstance(self.data, list):
            dict_list = []
            for i in self.data:
                if isinstance(i, PropertyTable):
                    dict_list.append(i.toDict())
                else:
                    dict_list.append(i.__dict__)
            pretty.prettyDataList(dict_list)

        if isinstance(self.data, Any):
            dict_list = self.data.__dict__
            pretty.prettyData(dict_list)


class JobResponse(BaseModel):

    '''
        Data class for JobResponse
    '''

    def __init__(self,jobId:str = None,statistics:Statistics = None,status:Status = None
                 ):
        self.jobId = jobId
        self.statistics = statistics
        self.status = status


class Response(BaseModel):
    '''
        Data class for Response.
    '''
    def __init__(self, status: Status = None,
                 items: Dict[str,DataItem] = None,
                 statistics: Statistics = None,
                 aliases: List[Alias] = None,
                 explainPlan: ExplainPlan = None):
        self.status = status
        self.items = items
        self.aliases = aliases
        self.statistics = statistics
        self.explainPlan = explainPlan

    def alias(self, alias: str) -> DataItem:
        if self.items == None:
            return DataItem(alias, None, ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET))
        if self.items.get(alias):
            return self.items.get(alias)
        return DataItem(alias, None, ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET))

    def aliasItem(self, alias: str) -> DataItem:
        if self.items == None:
            return DataItem(alias, None, ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET))
        if self.items.get(alias):
            return DataItem(alias, self.items[f'{alias}'].entities, ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE))
        return DataItem(alias, None, ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET))

    def get(self, index: int) -> DataItem:
        if len(self.aliases) - 1 >= index:
            data = self.items.get(self.aliases[index].alias)
            if data:
                return data
            if self.aliases[index].type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
                return DataItem(None, None, ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET))
        return DataItem(None, None, ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET))

    def getExplainPlan(self):
        return self.explainPlan

    def Print(self):
        pretty = PrettyPrint()
        pretty.prettyStatus(self.status)
        if self.status.code != ErrorCode.SUCCESS:
            return
        if self.statistics:
            pretty.prettyStatistics(self.statistics)
        explains = []
        if self.items:
            for key in self.items:
                dataItem = self.items.get(key)
                if dataItem.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_NODE):
                    pretty.prettyNode(dataItem)

                elif dataItem.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_EDGE):
                    pretty.prettyEdge(dataItem)

                elif dataItem.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
                    pretty.prettyTable(dataItem)

                elif dataItem.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_PATH):
                    pretty.prettyPath(dataItem)

                elif dataItem.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_ATTR):
                    pretty.prettyAttr(dataItem)

        if self.explainPlan:
            # for explan in self.explainPlan:
            #     explains.append(explan)
            # if explains:
            pretty.prettyTree(self.explainPlan)


class QueryResponseListener:
    def __init__(self):
        self._listeners = {}
        self._default_handlers = {
            "start": self._default_start_handler,
            "end": self._default_end_handler
        }

    def on(self, event: str, handler: Callable):
        if event not in self._listeners:
            self._listeners[event] = []
        self._listeners[event].append(handler)

    def emit(self, event: str, *args, **kwargs):
        if event in self._listeners:
            for handler in self._listeners[event]:
                handler(*args, **kwargs)
        elif event in self._default_handlers:
            self._default_handlers[event](*args, **kwargs)

    def _default_start_handler(self, request_config):
        print("Stream started with request:", request_config)

    def _default_end_handler(self, request_config):
        print("Stream ended with request:", request_config)


class ResponseCommon(UltipaResponse):
    data: None


# class User(BaseModel):
# 	username: str
# 	create: str
# 	last_login_time: str
# 	graphPrivileges: dict
# 	systemPrivileges: List[str]
# 	policies: List[str]


# class Index(BaseModel):
# 	name: str
# 	properties: str
# 	schema: str
# 	status: str
# 	size:str
# 	dbType:DBType


# class IndexTable:
# 	name: str
# 	data: List[Index]


class Stat(BaseModel):
    cpuUsage: str
    memUsage: str
    company: str
    cpuCores: str
    expiredDate: str
    serverType: str
    version: str

class Return_Type(BaseModel):
    is_realtime: bool
    is_visualization: bool
    is_wirte_back: bool


class Task_info(BaseModel):
    task_id: int = None
    server_id: int = None
    algo_name: str = None
    start_time: int = None
    writing_start_time: int = None
    end_time: int = None
    time_cost: int = None
    TASK_STATUS: int = None
    status_code: str = None
    engine_cost: int = None
    return_type: Return_Type = None


class Task(BaseModel):
    param: dict
    task_info: Task_info
    result: dict


class SearchKhop(BaseModel):
    values: dict
    nodes: List[Node]


class Path(BaseModel):
    nodes: List[Node]
    edges: List[Edge]


class SearchPath(BaseModel):
    paths: List[Path]


class NodeSpread(SearchPath):
    pass


class AutoNet(SearchPath):
    pass


class AlgoResultOpt(BaseModel):
    can_realtime: bool
    can_visualization: bool
    can_write_back: bool


# class Algo(BaseModel):
# 	param: dict
# 	name: str
# 	result_opt: AlgoResultOpt

class ResponseGraph(UltipaResponse):
    data: GraphSet


class ResponseListGraph(UltipaResponse):
    data: List[GraphSet]


# class ResponeListExta(UltipaResponse):
#     data: List[Exta]

class ResponseSchema(UltipaResponse):
    items: Schema


class ResponseListSchema(Response):
    items: List[Schema]


class ResponseListIndex(Response):
    data: List[Index]


# class ResponseListFulltextIndex(ResponseListIndex):
# 	pass


class ResponseSearchEdge(Response):
    items: List[DataItem]


class ResponseSearchNode(Response):
    items: List[DataItem]


class ResponseBulk:
    uuids: List[int]
    errorItem: Dict

class InsertResponse(BaseModel):

    '''
        Data class for InsertResponse

    '''

    def __init__(self,ids:List[str] = None,
                 uuids:List[int]= None,
                 errorItems: Dict[int,InsertErrorCode]= None,
                 statistics :Statistics= None,
                 status:Status= None
                 ):
        self.ids = ids
        self.uuids = uuids
        self.errorItems = errorItems
        self.statistics = statistics
        self.status = status

class ResponseInsertNode(Response):
    data: List[Node]


class ResponseInsertEdge(Response):
    data: List[Edge]


class ResponseDeleteNode(Response):
    data: List[Node]


class ResponseDeleteEdge(Response):
    data: List[Edge]


class ResponseBatchAutoInsert(UltipaResponse):
    data: Dict[str, ResponseBulk]


class ResponseListPolicy(UltipaResponse):
    data: List[Policy]


class ResponsePolicy(UltipaResponse):
    data: Policy


class ResponsePrivilege(UltipaResponse):
    data: List[Privilege]


class ResponseListProperty(Response):
    items: List[Property]


class ResponseProperty(UltipaResponse):
    data: List[Property]


class ResponseListTop(UltipaResponse):
    data: List[Top]


class ResponseListTask(UltipaResponse):
    data: List[Task]


class ResponseUser(UltipaResponse):
    data: User


class ResponseListUser(UltipaResponse):
    data: List[User]


class ResponseListAlgo(UltipaResponse):
    data: List[Algo]


class Cluster:
    host: str
    status: bool
    cpuUsage: str
    memUsage: str
    isLeader: bool
    isFollowerReadable: bool
    isAlgoExecutable: bool
    isUnset: bool


class ClusterInfo(UltipaResponse):
    data: List[Cluster]


class ResponseStat(UltipaResponse):
    data: Stats


class ResponseExport(UltipaResponse):
    data: List[Node]


class ResponseWithExistCheck:

    '''
        Data class for ResponseWithExistCheck

    '''

    def __init__(self,
                 exist:bool = None,
                 response: Response = None
                 ):
        self.exist = exist
        self.response = response


from ultipa.operate.base_extra import BaseExtra
from ultipa.types.types_response import *
from ultipa.structs.Index import Index
from ultipa.types.types import Status, ErrorCode
from ultipa.types.types_response import JobResponse
from ultipa.utils import UQLMAKER, CommandList
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.structs.DBType import DBType
from ultipa.utils.convert import convertTableToDict, convertToIndex, convertToFullText
from ultipa.utils.format import FormatType


class IndexExtra(BaseExtra):
    '''
	Processing class that defines settings for index related operations.
	'''

    def showIndex(self,
                  config: RequestConfig = RequestConfig()) -> List[Index]:
        '''
        Show all Index.

        Args:

            config: An object of RequestConfig class

        Returns:
            List[Index]
        '''
        command = CommandList.showIndex
        uqlMaker = UQLMAKER(command=command, commonParams=config)
        res = self.UqlListSimple(uqlMaker=uqlMaker, isSingleOne=False)
        indexdata = convertToIndex(res=res, all=True)
        if len(indexdata) > 0:
            res.data = indexdata
        else:
            res.data = None
        return res.data

    def showNodeIndex(self, config: RequestConfig = RequestConfig()) -> List[Index]:
        '''
        Show all Node index.

        Args:
            config: An object of RequestConfig class

        Returns:
            List[Index]
        '''
        command = CommandList.showNodeIndex
        uqlMaker = UQLMAKER(command=command, commonParams=config)
        res = self.UqlListSimple(uqlMaker=uqlMaker, isSingleOne=False)
        indexdata =  convertToIndex(res=res,all=False,dbtype=DBType.DBNODE)
        if len(indexdata) > 0:
            res.data = indexdata
        else:
            res.data = None
        return res.data

    def showEdgeIndex(self, config: RequestConfig = RequestConfig()) -> List[Index]:
        '''
        Show all Edge index.

        Args:
            config: An object of RequestConfig class

        Returns:
            List[Index]
        '''

        command = CommandList.showEdgeIndex
        uqlMaker = UQLMAKER(command=command, commonParams=config)
        res = self.UqlListSimple(uqlMaker=uqlMaker, isSingleOne=False)
        indexdata = convertToIndex(res=res, all=False, dbtype=DBType.DBEDGE)
        if len(indexdata) > 0:
            res.data = indexdata
        else:
            res.data = None
        return res.data

    def showFulltext(self,
                     config: RequestConfig = RequestConfig()) -> List[Index]:
        '''
        Show all full-text index.

        Args:
            config: An object of RequestConfig class
        Returns:

            List[Index]
        '''
        command = CommandList.showFulltext
        uqlMaker = UQLMAKER(command=command, commonParams=config)
        res = self.UqlListSimple(uqlMaker=uqlMaker, isSingleOne=False)
        fulltextdata = convertToFullText(res=res)
        if len(fulltextdata) > 0:
            res.data = fulltextdata
        else:
            res.data = None
        return res.data

    def createIndex(self, dbType: DBType, source: str, indexName: str,
                    config: RequestConfig = RequestConfig()) -> JobResponse:
        '''
        Create an index.

        Args:
            dbType: The DBType of data (DBNODE or DBEDGE)

            indexName: The name of index

            source: Property Related Information

            config: An object of RequestConfig class

        Returns:
            JobResponse
        '''
        response = Response()
        if FormatType.checkNone(dbType) or FormatType.checkNone(source) or FormatType.checkNone(indexName) or dbType == DBType.DBGLOBAL:
            message = 'The dbType cannot be None' if FormatType.checkNone(dbType) \
                else 'The source cannot be None' if FormatType.checkNone(source) else 'The indexName cannot be None' if FormatType.checkNone(source) else 'The dbType cannot be DBTYPE.DBGLOBAL'

            response.status=Status(code=ErrorCode.FAILED, message=message)

            return response
        command = dbType == DBType.DBNODE and CommandList.createNodeIndex or CommandList.createEdgeIndex
        uqlMaker = UQLMAKER(command=command, commonParams=config)
        source = f'{source}'
        uqlMaker.setCommandParams([source, indexName])
        res = self.uqlSingle(uqlMaker=uqlMaker)
        result = [{'new_job_id': ''}]
        if res.status.code == ErrorCode.SUCCESS and res.items:
            result = convertTableToDict(res.alias(res.aliases[0].alias).entities.rows,
                                        res.alias(res.aliases[0].alias).entities.headers)

        return JobResponse(jobId=result[0].get('new_job_id'), status=res.status, statistics=res.statistics)

    def createNodeIndex(self, source: str, indexName: str,
                        config: RequestConfig = RequestConfig()) -> JobResponse:
        '''
        Create an node index.

        Args:
            source: Property Related Information

            indexName: The name of index

            config: An object of RequestConfig class

        Returns:
            JobResponse
        '''
        return self.createIndex(dbType=DBType.DBNODE, source=source, indexName=indexName, config=config)

    def createEdgeIndex(self, source: str, indexName: str,
                        config: RequestConfig = RequestConfig()) -> JobResponse:
        '''
        Create an edge index.

        Args:
            source: Property Related Information

            indexName: The name of index

            config: An object of RequestConfig class

        Returns:
            JobResponse
        '''

        return self.createIndex(dbType=DBType.DBEDGE, source=source, indexName=indexName, config=config)

    def createFulltext(self, dbType: DBType, schemaName: str, propertyName: str, indexName: str,
                       config: RequestConfig = RequestConfig()) -> JobResponse:
        '''
        Create an full-text index.

        Args:
            dbType: The DBType of data (DBNODE or DBEDGE)

            schemaName: The name of schema

            propertyName: An object of propertyName

            indexName: Name of the fulltext index

            config: An object of RequestConfig class

        Returns:
            JobResponse

        '''
        command = dbType == DBType.DBNODE and CommandList.createNodeFulltext or CommandList.createEdgeFulltext
        command1 = "@`%s`.`%s`" % (schemaName, propertyName)
        commandP = [command1, indexName]
        uqlMaker = UQLMAKER(command=command, commonParams=config)
        uqlMaker.setCommandParams(commandP=commandP)
        res = self.uqlSingle(uqlMaker=uqlMaker)
        result = [{'new_job_id': ''}]
        if res.status.code == ErrorCode.SUCCESS:
            result = convertTableToDict(res.alias('result').entities.rows, res.alias('result').entities.headers)

        return JobResponse(jobId=result[0].get('new_job_id'), status=res.status, statistics=res.statistics)

    def createNodeFulltext(self, schemaName: str, propertyName: str, indexName: str,
                           config: RequestConfig = RequestConfig()) -> JobResponse:
        '''
        Create an node full-text index.

        Args:

            schemaName: The name of schema

            propertyName: An name of property

            indexName: Name of the fulltext index

            config: An object of RequestConfig class

        Returns:
            JobResponse

        '''
        return self.createFulltext(dbType=DBType.DBNODE, schemaName=schemaName, propertyName=propertyName,
                                   indexName=indexName, config=config)

    def createEdgeFulltext(self, schemaName: str, propertyName: str, indexName: str,
                           config: RequestConfig = RequestConfig()) -> JobResponse:
        '''
        Create an edge full-text index.

        Args:

            schemaName: The name of schema

            propertyName: An name of property

            indexName: Name of the fulltext index

            config: An object of RequestConfig class

        Returns:
            JobResponse

		'''
        return  self.createFulltext(dbType=DBType.DBEDGE, schemaName=schemaName, propertyName=propertyName,
                                  indexName=indexName, config=config)

    def dropIndex(self, dbType: DBType, indexName: str,
                  config: RequestConfig = RequestConfig()) -> Response:
        '''
        Drop an index.

        Args:
            dbType: The DBType of data (DBNODE or DBEDGE)

            indexName: The name of index

            config: An object of RequestConfig class

        Returns:
            Response
        '''
        command = dbType == DBType.DBNODE and CommandList.dropNodeIndex or CommandList.dropEdgeIndex
        commandP = f'"{indexName}"'

        uqlMaker = UQLMAKER(command=command, commandP=commandP, commonParams=config)
        res = self.uqlSingle(uqlMaker=uqlMaker)
        return res

    def dropNodeIndex(self, indexName: str,
                      config: RequestConfig = RequestConfig()) -> Response:
        '''
        Drop an node index.

        Args:
            indexName: The name of index

            config: An object of RequestConfig class

        Returns:
            Response
        '''
        return self.dropIndex(dbType=DBType.DBNODE, indexName=indexName, config=config)

    def dropEdgeIndex(self, indexName: str,
                      config: RequestConfig = RequestConfig()) -> Response:
        '''
        Drop an edge index.

        Args:
            indexName: The name of index

            config: An object of RequestConfig class

        Returns:
            Response
        '''

        return self.dropIndex(dbType=DBType.DBEDGE, indexName=indexName, config=config)

    def dropFulltext(self, dbType: DBType, fulltextName: str,
                     config: RequestConfig = RequestConfig()) -> Response:
        '''
        Drop an full-text index.

        Args:
            dbType: The DBType of data (DBNODE or DBEDGE)

            fulltextName: The name of the fulltext index

            config: An object of RequestConfig class

        Returns:
            Response
        '''

        command = dbType == DBType.DBNODE and CommandList.dropNodeFulltext or CommandList.dropEdgeFulltext
        uqlMaker = UQLMAKER(command=command, commonParams=config)
        uqlMaker.setCommandParams(fulltextName)
        res = self.uqlSingle(uqlMaker=uqlMaker)
        return res

    def showNodeFulltext(self, config: RequestConfig = RequestConfig()) -> List[Index]:
        '''
        Show all full-text Node index.

        Args:

            config: An object of RequestConfig class

        Returns:

            List[Index]
        '''
        command = CommandList.showNodeFulltext
        uqlMaker = UQLMAKER(command=command, commonParams=config)
        res = self.UqlListSimple(uqlMaker=uqlMaker, isSingleOne=False)
        fulltextdata = convertToFullText(res=res, all=False, dbtype=DBType.DBNODE)
        if len(fulltextdata) > 0:
            res.data = fulltextdata
        else:
            res.data = None
        return res.data

    def showEdgeFulltext(self, config: RequestConfig = RequestConfig()) -> List[Index]:
        '''
        Show all full-text Edge index.

        Args:

            config: An object of RequestConfig class

        Returns:

            List[Index]
        '''
        command = CommandList.showEdgeFulltext
        uqlMaker = UQLMAKER(command=command, commonParams=config)
        res = self.UqlListSimple(uqlMaker=uqlMaker, isSingleOne=False)
        fulltextdata = convertToFullText(res=res, all=False, dbtype=DBType.DBEDGE)
        if len(fulltextdata) > 0:
            res.data = fulltextdata
        else:
            res.data = None
        return res.data



# -*- coding: utf-8 -*-
# @Time    : 2023/8/4 16:19
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Edge.py
from typing import Dict

from ultipa.structs.BaseModel import BaseModel


class Edge(BaseModel):
    '''
        Data class for edge.
    '''
    _index = None

    def __init__(self,
                 uuid: int = None,
                 fromId: str = None,
                 toId: str = None,
                 fromUuid: int = None,
                 toUuid: int = None,
                 schema: str = None,
                 values: Dict[str,any] = None):
        self.schema = schema
        self.uuid = uuid
        self.fromId = fromId
        self.toId = toId
        self.fromUuid = fromUuid
        self.toUuid = toUuid
        self.values = values

    def getFrom(self):
        return self.fromId

    def getTo(self):
        return self.toId

    def getFromUUID(self):
        return self.fromUuid

    def getUUID(self):
        return self.uuid

    def getToUUID(self):
        return self.toUuid

    def getValues(self):
        return self.values

    def getSchema(self):
        return self.schema

    def get(self, propName: str):
        return self.values.get(propName)

    def set(self, propName: str, value):
        self.values.update({propName: value})

    def _getIndex(self):
        return self._index


# -*- coding: utf-8 -*-
# @Time    : 2023/8/4 16:18
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Node.py
from typing import Dict, List

from ultipa.structs.BaseModel import BaseModel


class Node(BaseModel):
    '''
        Data calss for node.
    '''
    _index = None

    # **kwargs
    def __init__(self,
                 uuid: int = None,
                 id: str = None,
                 schema: str = None,
                 values: Dict[str,any] = None
                 ):
        self.uuid = uuid
        self.id = id
        self.schema = schema
        self.values = values

    def getID(self):
        return self.id

    def getUUID(self):
        return self.uuid

    def getValues(self):
        return self.values

    def getSchema(self):
        return self.schema

    def get(self, propName: str):
        return self.values.get(propName)

    def set(self, propName: str, value):
        self.values.update({propName: value})

    def _getIndex(self):
        return self._index


class NodeAlias:
    def __init__(self, alias: str, nodes: List[Node] = None):
        self.alias = alias
        if nodes is None:
            nodes = []
        self.nodes = nodes


from ultipa.structs.BaseModel import BaseModel

class Projection(BaseModel):
    '''

        Data class for Projection

    '''
    def __init__(self,name: str = None,graphName: str = None,status: str =None,stats: str = None,config: str = None):
        self.name = name
        self.graphName = graphName
        self.status = status
        self.stats = stats
        self.config = config


# -*- coding: utf-8 -*-
# @Time    : 2024/05/14 11:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Policy.py
from typing import List, Dict
from ultipa.structs.BaseModel import BaseModel
from ultipa.structs.PropertyPrivilege import PropertyPrivilege


class Policy(BaseModel):
    '''
        Data class for Policy.
    '''

    def __init__(self, name: str = None,
                 graphPrivileges: Dict[str,List[str]] = None,
                 systemPrivileges: List[str] = None,
                 propertyPrivileges: PropertyPrivilege = None,
                 policies: List[str] = None):
        self.name = name
        self.systemPrivileges = systemPrivileges
        self.graphPrivileges = graphPrivileges
        self.propertyPrivileges = propertyPrivileges
        self.policies = policies



from ultipa.structs.BaseModel import BaseModel

class License(BaseModel):
    def __init__(self, limitedShard: str = '', limitedHdc: str = '', expiredDate: str = '',
                 licenseUUId: str = '', company: str = '', department: str = '', limitedUser: str = '',
                 limitedGraph: str = '', limitedNode: str = '',
                 limitedEdge: str = ''):
        # self.licenseUUId = licenseUUId
        # self.company = company
        # self.department = department
        # self.limitedUser = limitedUser
        # self.limitedGraph = limitedGraph
        # self.limitedNode = limitedNode
        # self.limitedEdge = limitedEdge
        self.limitedShard = limitedShard
        self.limitedHdc = limitedHdc
        self.expiredDate = expiredDate


import ast
import json
import datetime
from typing import List

from ultipa.structs.Job import Job
from ultipa.structs.License import License

from ultipa.structs.BaseModel import BaseModel
from ultipa.structs.DBType import DBType
from ultipa.structs.Index import Index
from ultipa.structs.Process import Process
from ultipa.structs.Projection import Projection
from ultipa.structs.PropertyPrivilege import PropertyPrivilege, PropertyPrivilegeElement
from ultipa.structs.Schema import SchemaStat
from ultipa.structs.User import User
from ultipa.structs.Policy import Policy
from ultipa.structs.Privilege import Privilege, PrivilegeLevel
from ultipa.structs.GraphSet import GraphSet
from ultipa.structs.Stats import Stats
from ultipa.structs.Top import Top
from ultipa.structs.Algo import Algo
from ultipa.structs.HDC import HDCGraph
from ultipa.utils.noneCheck import checkNone


class Any(BaseModel):
    '''
    Any model
    '''

    def __str__(self):
        return str(self.__dict__)

    pass


def convertToAnyObject(dict1: dict):
    '''
    Convert Dict to Object.

    Args:
        dict1:

    Returns:

    '''
    obj = Any()
    for k in dict1.keys():
        v = dict1[k]
        if isinstance(v, list):
            for i, n in enumerate(v):
                if isinstance(n, dict):
                    v[i] = convertToAnyObject(n)
        # if isinstance(v, dict):
        #     v = convertToAnyObject(v)
        obj.__setattr__(k, v)
    return obj


def convertToTask(dict1: dict):
    '''
    Convert Dict to Object.

    Args:
        dict1:

    Returns:

    '''

    from ultipa.types.types_response import Task_info
    taskInfo = Task_info()
    for k, value in dict1.items():
        # if isinstance(v, dict):
        #     v = convertToAnyObject(v)
        if hasattr(taskInfo, k):
            setattr(taskInfo, k, value)

    return taskInfo


def convertToListAnyObject(list1: List[dict]):
    '''
    Convert List[Dict] to Object.

    Args:
        list1:

    Returns:

    '''
    if not list1 and isinstance(list1, list):
        return list1
    if not list1:
        return
    newList = []
    for dict1 in list1:
        newList.append(convertToAnyObject(dict1))
    return newList


def convertTableToDict(table_rows, headers):
    '''
    Convert Table to Object.

    Args:
        table_rows:
        headers:

    Returns:

    '''
    newList = []
    for data in table_rows:
        dic = {}
        for index, header in enumerate(headers):
            dic.update({header.get("property_name"): data[index]})
        newList.append(dic)
    return newList


def convertToGraph(res):
    if isinstance(res.data, list) and len(res.data) > 1:
        graphdata = res.data
        graphdata.sort(key=lambda graph: int(graph.id))
        graphlist = []
        for graph in graphdata:
            graphlist.append(GraphSet(
                id=graph.id,
                name=graph.name,
                totalNodes=int(graph.total_nodes) if hasattr(graph,
                                                             'total_nodes') and graph.total_nodes.isdigit() else 0,
                totalEdges=int(graph.total_edges) if hasattr(graph,
                                                             'total_edges') and graph.total_edges.isdigit() else 0,
                description=graph.description,
                status=graph.status,
                shards=graph.shards.split(','),
                partitionBy=graph.partition_by,
                slotNum=int(graph.slot_num) if graph.slot_num.isdigit else 0
            ))

        return graphlist
    else:
        graph = res.data[0]
        return GraphSet(
            id=graph.id,
            name=graph.name,
            totalNodes=int(graph.total_nodes) if hasattr(graph, 'total_nodes') and graph.total_nodes.isdigit() else 0,
            totalEdges=int(graph.total_edges) if hasattr(graph, 'total_edges') and graph.total_edges.isdigit() else 0,
            description=graph.description,
            status=graph.status,
            shards=graph.shards.split(','),
            partitionBy=graph.partition_by,
            slotNum=int(graph.slot_num) if graph.slot_num.isdigit else 0
        )


def convertToIndex(res, all: bool = True, dbtype: DBType = None):
    nodeindex = []
    edgeindex = []
    indexdata = []
    if all:
        nodeindex = res.data[0].data
        edgeindex = res.data[1].data
    elif dbtype == DBType.DBNODE:
        nodeindex = res.data[0].data
    elif dbtype == DBType.DBEDGE:
        edgeindex = res.data[0].data

    if nodeindex:
        for index in nodeindex:
            indexdata.append(Index(
                id=index.id,
                name=index.name,
                properties=index.properties,
                schema=index.schema,
                # size=index.size,
                status=index.status,
                dbType=DBType.DBNODE
            ))
    if edgeindex:
        for index in edgeindex:
            indexdata.append(Index(
                id=index.id,
                name=index.name,
                properties=index.properties,
                schema=index.schema,
                # size=index.size,
                status=index.status,
                dbType=DBType.DBEDGE
            ))
    return indexdata


def convertToFullText(res, all: bool = True, dbtype: DBType = None):
    nodeindex = []
    edgeindex = []
    indexdata = []
    if all:
        nodeindex = res.data[0].data
        edgeindex = res.data[1].data
    elif dbtype == DBType.DBNODE:
        nodeindex = res.data[0].data
    elif dbtype == DBType.DBEDGE:
        edgeindex = res.data[0].data

    if nodeindex:
        for index in nodeindex:
            indexdata.append(Index(
                name=index.name,
                properties=index.properties,
                schema=index.schema,
                status=index.status,
                dbType=DBType.DBNODE
            ))
    if edgeindex:
        for index in edgeindex:
            indexdata.append(Index(
                name=index.name,
                properties=index.properties,
                schema=index.schema,
                status=index.status,
                dbType=DBType.DBEDGE
            ))
    return indexdata


def convertToUser(res):
    if isinstance(res.data, list) and len(res.data) > 1:
        userdata = res.data
        datalist = []
        for data in userdata:
            proPrivilegesData = json.loads(data.propertyPrivileges)
            datalist.append(User(
                username=data.username,
                systemPrivileges=data.systemPrivileges,
                graphPrivileges=data.graphPrivileges,
                createdTime=datetime.datetime.fromtimestamp(data.create),
                propertyPrivileges=PropertyPrivilege(
                    PropertyPrivilegeElement(
                        proPrivilegesData['node'].get('read'),
                        proPrivilegesData['node'].get('write'),
                        proPrivilegesData['node'].get('deny'),
                    ),
                    PropertyPrivilegeElement(
                        proPrivilegesData['edge'].get('read'),
                        proPrivilegesData['edge'].get('write'),
                        proPrivilegesData['edge'].get('deny'),
                    )
                ),
                policies=data.policies
            ))

        return datalist
    else:
        data = res.data[0]
        proPrivilegesData = json.loads(data.propertyPrivileges)
        return User(
            username=data.username,
            systemPrivileges=data.systemPrivileges,
            graphPrivileges=data.graphPrivileges,
            createdTime=datetime.datetime.fromtimestamp(data.create),
            propertyPrivileges=PropertyPrivilege(
                PropertyPrivilegeElement(
                    proPrivilegesData['node'].get('read'),
                    proPrivilegesData['node'].get('write'),
                    proPrivilegesData['node'].get('deny'),
                ),
                PropertyPrivilegeElement(
                    proPrivilegesData['edge'].get('read'),
                    proPrivilegesData['edge'].get('write'),
                    proPrivilegesData['edge'].get('deny'),
                )
            ),
            policies=data.policies
        )


def convertToPolicy(res):
    if isinstance(res.data, list) and len(res.data) > 1:
        policydata = res.data
        policylist = []
        for data in policydata:
            proPrivilegesData = json.loads(data.propertyPrivileges)
            policylist.append(Policy(
                name=data.name,
                systemPrivileges=data.systemPrivileges,
                graphPrivileges=data.graphPrivileges,
                policies=data.policies,
                propertyPrivileges=PropertyPrivilege(
                    PropertyPrivilegeElement(
                        proPrivilegesData['node'].get('read'),
                        proPrivilegesData['node'].get('write'),
                        proPrivilegesData['node'].get('deny'),
                    ),
                    PropertyPrivilegeElement(
                        proPrivilegesData['edge'].get('read'),
                        proPrivilegesData['edge'].get('write'),
                        proPrivilegesData['edge'].get('deny'),
                    )
                )
            ))

        return policylist
    else:
        data = res.data[0]
        proPrivilegesData = json.loads(data.propertyPrivileges)
        return Policy(
            name=data.name,
            systemPrivileges=data.systemPrivileges,
            graphPrivileges=data.graphPrivileges,
            policies=data.policies,
            propertyPrivileges=PropertyPrivilege(
                PropertyPrivilegeElement(
                    proPrivilegesData['node'].get('read'),
                    proPrivilegesData['node'].get('write'),
                    proPrivilegesData['node'].get('deny'),
                ),
                PropertyPrivilegeElement(
                    proPrivilegesData['edge'].get('read'),
                    proPrivilegesData['edge'].get('write'),
                    proPrivilegesData['edge'].get('deny'),
                )
            )
        )


def convertToPrivilege(res):
    if isinstance(res.data, list) and len(res.data) > 1:
        privilegedata = res.data
        privilegelist = []
        for data in privilegedata:
            sysPrivileges = [Privilege(name=sysdata, level=PrivilegeLevel.SYSTEM) for sysdata in data.systemPrivileges]
            privilegelist.extend(sysPrivileges)
            graPrivileges = [Privilege(name=gradata, level=PrivilegeLevel.GRAPH) for gradata in data.graphPrivileges]
            privilegelist.extend(graPrivileges)

        return privilegelist
    else:
        data = res.data[0]
        sysPrivileges = [Privilege(name=sysdata, level=PrivilegeLevel.SYSTEM) for sysdata in data.systemPrivileges]
        graPrivileges = [Privilege(name=gradata, level=PrivilegeLevel.GRAPH) for gradata in data.graphPrivileges]
        sysPrivileges.extend(graPrivileges)
        return sysPrivileges


def convertToStats(res):
    statsdata = res.data[0]
    return Stats(
        limitedHdc=statsdata.limited_hdc,
        limitedShard=statsdata.limited_shard,
        expiredDate=statsdata.expired_date
    )


def convertToSchemas(res):
    from ultipa.types.types import Schemas
    graphcount = convertTableToDict(res.alias("_graphCount").entities.rows, res.alias("_graphCount").entities.headers)
    schemasresult = Schemas()
    count = {}
    for data in graphcount:
        if not checkNone(data.get('schema')):
            edgepair = count.get(data.get('schema')) if count.get(data.get('schema')) is not None else []
            data['count'] = int(data.get('count')) if data.get('count') and data.get(
                                         'count').isdigit() else 0
            edgepair.append(data)
            count[data.get('schema')] = edgepair
    schemas = []
    for aliases in res.aliases:
        nodeschemas = res.alias("_nodeSchema").asSchemas() if aliases.alias == '_nodeSchema' else []
        schemas.extend(nodeschemas)
        edgeschemas = res.alias("_edgeSchema").asSchemas() if aliases.alias == '_edgeSchema' else []
        schemas.extend(edgeschemas)

    for data in schemas:
        if data.name in count:
            countsum = 0
            for countiter in count.get(data.name):
                countsum += countiter.get('count')
            data.total = countsum
            data.stats = [SchemaStat(fromSchema=pairiter.get('from_schema'),
                                     toSchema=pairiter.get('to_schema'),
                                     count=pairiter.get('count'),
                                     type=DBType.DBNODE if pairiter.get('type') in 'node' else DBType.DBEDGE  if pairiter.get('type') in 'edge' else None ,
                                     name=pairiter.get('schema')
                                     )
                          for pairiter in count.get(data.name)]

    schemasresult.schema = schemas
    return schemasresult


def convertToTop(res) -> List[Process]:
    if isinstance(res.data, list) and len(res.data) > 0:
        topdata = res.data
        toplist = []
        for data in topdata:
            toplist.append(Process(
                processId=data.process_id,
                processQuery=data.process_query,
                duration=data.duration,
                status=data.status
            ))

    return toplist


def convertToAlgo(res):
    algodata = res.data
    algolist = []
    for data in algodata:
        desc = json.loads(data.get("description"))
        algolist.append(Algo(
            name=data.name,
            version=desc.param['version'],
            params=data.param['parameters'],
            description=data.get("description"),
            type=data.param['type'],
            writeSupportType=data.param['write_support_type'],
            canRollback=data.param['can_rollback'],
            configContext=data.param['config_context']
        ))

    if len(algolist) == 1:
        return algolist[0]
    else:
        return algolist


def convertToExta(res):
    from ultipa.types.types import Exta
    extadata = res.data
    extalist = []
    for data in extadata:
        extalist.append(Exta(author=data.author,
                             detail=data.detail,
                             name=data.name,
                             version=data.version))

    return extalist


def convertToLicense(res):
    return License(limitedHdc=res.get('limited_hdc'),
                   limitedShard=res.get('limited_shard'),
                   expiredDate=res.get('expired_date'))


def converToJob(obj):
    return Job(id=obj.get('job_id'), graphName=obj.get('graph_name'), type=obj.get('type'),
               query=obj.get('query'), status=obj.get('status'), errMsg=obj.get('err_msg'),
               result=json.loads(obj.get('result')) if obj.get('result') != '' else {}, startTime=obj.get('start_time'),
               endTime=obj.get('end_time'), progress=obj.get('progress'))


def converToProjection(res):
    if isinstance(res, list) and len(res) > 0:
        return [Projection(name=data.get('name'), graphName=data.get('graph_name'), status=data.get('status'),
                           stats=data.get('stats'), config=data.get('config')) for data in res]
    return res


def converToHDCGraph(res):
    if isinstance(res, list) and len(res) > 0:
        return [HDCGraph(name=data.get('name'),
                         graphName=data.get('graph_name'),
                         status=data.get('status'),
                         stats=data.get('stats'),
                         isDefault=data.get('is_default'),
                         hdcServerName=data.get('hdc_server_name'),
                         hdcServerStatus=data.get('hdc_server_status'),
                         config=data.get('config')) for data in res]
    return res
